"""Tests for context propagation utilities."""

from __future__ import annotations

import logging
import os
import threading
from collections.abc import Iterator, Mapping, MutableMapping
from unittest.mock import patch

import pytest
from inline_snapshot import snapshot
from opentelemetry import propagate, trace
from opentelemetry.context import Context
from opentelemetry.sdk.trace import TracerProvider

from mirascope.ops import (
    SESSION_HEADER_NAME,
    ContextPropagator,
    PropagatorFormat,
    extract_context,
    extract_session_id,
    get_propagator,
    inject_context,
    propagated_context,
    reset_propagator,
    session,
)
from mirascope.ops._internal.propagation import (
    ENV_PROPAGATOR_FORMAT,
    ENV_PROPAGATOR_SET_GLOBAL,
)
from mirascope.ops.exceptions import ConfigurationError
from tests.ops.conftest import (
    PROPAGATOR_FORMATS,
    TRACEPARENT_PATTERN,
    VALID_SPAN_ID,
    VALID_TRACE_ID,
    VALID_TRACEPARENT,
)

PROPAGATOR_FORMAT_CONFIGS = [
    (None, {"traceparent"}),
    ("tracecontext", {"traceparent"}),
    ("b3", {"b3"}),
    ("b3multi", {"x-b3-traceid", "x-b3-spanid", "x-b3-sampled"}),
    ("jaeger", {"uber-trace-id"}),
    (
        "composite",
        {
            "traceparent",
            "b3",
            "x-b3-traceid",
            "x-b3-spanid",
            "x-b3-sampled",
            "uber-trace-id",
        },
    ),
]


class InvalidCarrier(Mapping[str, str]):
    """Mock carrier that raises errors during extraction."""

    def __getitem__(self, key: str) -> str:
        raise RuntimeError("extraction error")

    def __iter__(self) -> Iterator[str]:
        raise RuntimeError("extraction error")

    def __len__(self) -> int:
        raise RuntimeError("extraction error")


class ImmutableCarrier(MutableMapping[str, str]):
    """Mock carrier that raises errors during injection."""

    def __init__(self) -> None:
        self._data: dict[str, str] = {}

    def __getitem__(self, key: str) -> str:
        return self._data[key]

    def __setitem__(self, key: str, value: str) -> None:
        raise RuntimeError("injection error")

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)


def _configure_propagator(format_name: str) -> None:
    """Configure propagator format and reset singleton."""
    os.environ[ENV_PROPAGATOR_FORMAT] = format_name
    reset_propagator()


@pytest.mark.parametrize(
    ("propagator_format", "expected_headers"),
    PROPAGATOR_FORMAT_CONFIGS,
    ids=["default"] + PROPAGATOR_FORMATS,
)
def test_propagator_formats(
    propagator_format: PropagatorFormat | None,
    expected_headers: set[str],
    tracer_provider: TracerProvider,
) -> None:
    """Create correct propagator based on MIRASCOPE_PROPAGATOR environment variable."""
    if propagator_format is None:
        os.environ.pop(ENV_PROPAGATOR_FORMAT, None)
    else:
        os.environ[ENV_PROPAGATOR_FORMAT] = propagator_format

    context_propagator = ContextPropagator(set_global=False)

    tracer = tracer_provider.get_tracer(__name__)
    with tracer.start_as_current_span("test-span") as span:
        current_context = trace.set_span_in_context(span)
        carrier: dict[str, str] = {}
        context_propagator.inject_context(carrier, context=current_context)

        actual_headers = set(carrier)
        missing = expected_headers - actual_headers
        unexpected = actual_headers - expected_headers
        assert not missing, f"Missing headers: {missing}"
        assert not unexpected, f"Unexpected headers: {unexpected}"


def test_propagator_invalid_format_raises_error(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Raise ConfigurationError and log error for invalid propagator format."""
    os.environ[ENV_PROPAGATOR_FORMAT] = "invalid_format"

    with (
        caplog.at_level(logging.ERROR),
        pytest.raises(ConfigurationError) as exception_info,
    ):
        ContextPropagator()

    assert str(exception_info.value) == snapshot(
        "Invalid propagator format: invalid_format. "
        "Valid options: tracecontext, b3, b3multi, jaeger, composite"
    )
    assert caplog.record_tuples == snapshot(
        [
            (
                "mirascope.ops._internal.propagation",
                logging.ERROR,
                "Invalid propagator format: invalid_format. "
                "Valid options: tracecontext, b3, b3multi, jaeger, composite",
            )
        ]
    )


def test_propagator_sets_global_by_default() -> None:
    """Set the propagator as global textmap propagator by default."""
    os.environ[ENV_PROPAGATOR_FORMAT] = "tracecontext"
    propagator = ContextPropagator()
    assert propagate.get_global_textmap() is propagator._propagator  # pyright: ignore[reportPrivateUsage]


def test_propagator_respects_set_global_false() -> None:
    """Do not set global propagator when set_global=False."""
    os.environ[ENV_PROPAGATOR_FORMAT] = "tracecontext"
    previous_global = propagate.get_global_textmap()
    ContextPropagator(set_global=False)
    assert propagate.get_global_textmap() is previous_global


def test_propagator_respects_set_global_guard() -> None:
    """Do not set global propagator when _MIRASCOPE_PROPAGATOR_SET_GLOBAL is false."""
    os.environ[ENV_PROPAGATOR_FORMAT] = "tracecontext"
    os.environ[ENV_PROPAGATOR_SET_GLOBAL] = "false"
    previous_global = propagate.get_global_textmap()
    ContextPropagator()
    assert propagate.get_global_textmap() is previous_global


@pytest.mark.parametrize(
    ("carrier", "expected"),
    [
        pytest.param(
            {"traceparent": VALID_TRACEPARENT},
            {
                "is_valid": True,
                "trace_id": VALID_TRACE_ID,
                "span_id": VALID_SPAN_ID,
            },
            id="valid_headers",
        ),
        pytest.param(
            {"traceparent": "invalid-traceparent-format"},
            {"is_valid": False},
            id="invalid_headers",
        ),
        pytest.param(
            {},
            {"is_valid": False},
            id="missing_headers",
        ),
    ],
)
def test_propagator_extract_context(
    carrier: dict[str, str],
    expected: dict[str, bool | int],
) -> None:
    """Extract context from carrier with valid, invalid, or missing headers."""
    os.environ[ENV_PROPAGATOR_FORMAT] = "tracecontext"

    context_propagator = get_propagator()

    extracted_context = context_propagator.extract_context(carrier)

    assert isinstance(extracted_context, Context)

    extracted_span = trace.get_current_span(extracted_context)
    span_context = extracted_span.get_span_context()

    result: dict[str, bool | int] = {"is_valid": span_context.is_valid}
    if span_context.is_valid:
        result["trace_id"] = span_context.trace_id
        result["span_id"] = span_context.span_id

    assert result == expected


def test_propagator_handles_extraction_errors(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Log debug message when extraction fails on malformed carrier."""
    os.environ[ENV_PROPAGATOR_FORMAT] = "tracecontext"

    context_propagator = get_propagator()

    with caplog.at_level(logging.DEBUG):
        carrier = InvalidCarrier()
        result = context_propagator.extract_context(carrier)

        assert caplog.record_tuples == snapshot(
            [
                (
                    "mirascope.ops._internal.propagation",
                    logging.DEBUG,
                    "Failed to extract context from carrier: "
                    "RuntimeError: extraction error",
                )
            ]
        )
        assert isinstance(result, Context)


def test_propagator_injects_context_into_carrier(
    tracer_provider: TracerProvider,
) -> None:
    """Inject current context into carrier headers."""
    os.environ[ENV_PROPAGATOR_FORMAT] = "tracecontext"

    tracer = tracer_provider.get_tracer(__name__)

    context_propagator = get_propagator()

    with tracer.start_as_current_span("test-span") as span:
        current_context = trace.set_span_in_context(span)
        carrier: dict[str, str] = {}
        context_propagator.inject_context(carrier, context=current_context)

        assert set(carrier.keys()) == snapshot({"traceparent"})
        assert TRACEPARENT_PATTERN.match(carrier["traceparent"])


@pytest.mark.parametrize(
    "use_method",
    [
        pytest.param(True, id="method"),
        pytest.param(False, id="convenience_function"),
    ],
)
def test_inject_specific_context(
    tracer_provider: TracerProvider,
    use_method: bool,
) -> None:
    """Inject a specific context rather than current context."""
    os.environ[ENV_PROPAGATOR_FORMAT] = "tracecontext"

    tracer = tracer_provider.get_tracer(__name__)
    context_propagator = get_propagator()

    with tracer.start_as_current_span("test-span") as span:
        specific_context = Context()
        specific_context = trace.set_span_in_context(span, specific_context)

        carrier: dict[str, str] = {}

        if use_method:
            context_propagator.inject_context(carrier, context=specific_context)
        else:
            inject_context(carrier, context=specific_context)

        assert set(carrier.keys()) == snapshot({"traceparent"})


def test_propagator_handles_injection_errors(
    caplog: pytest.LogCaptureFixture,
    tracer_provider: TracerProvider,
) -> None:
    """Log debug message when injection fails on immutable carrier."""
    os.environ[ENV_PROPAGATOR_FORMAT] = "tracecontext"

    context_propagator = get_propagator()
    tracer = tracer_provider.get_tracer(__name__)

    with tracer.start_as_current_span("test-span") as span:
        current_context = trace.set_span_in_context(span)

        with caplog.at_level(logging.DEBUG):
            carrier = ImmutableCarrier()
            context_propagator.inject_context(carrier, context=current_context)

            assert caplog.record_tuples == snapshot(
                [
                    (
                        "mirascope.ops._internal.propagation",
                        logging.DEBUG,
                        "Failed to inject context into carrier: "
                        "RuntimeError: injection error",
                    )
                ]
            )


def test_get_propagator_singleton() -> None:
    """Return the same ContextPropagator instance on multiple calls."""
    first_call = get_propagator()
    second_call = get_propagator()

    assert first_call is second_call
    assert isinstance(first_call, ContextPropagator)


def test_get_propagator_creates_instance_once() -> None:
    """Create propagator instance only on first call to get_propagator."""
    reset_propagator()

    with patch(
        "mirascope.ops._internal.propagation.ContextPropagator", wraps=ContextPropagator
    ) as mock_constructor:
        get_propagator()
        get_propagator()
        get_propagator()

        assert mock_constructor.call_count == 1


def test_extract_context_convenience_function(
    valid_carrier: dict[str, str],
) -> None:
    """Module-level extract_context delegates to propagator instance."""

    extracted_context = extract_context(valid_carrier)

    assert isinstance(extracted_context, Context)

    extracted_span = trace.get_current_span(extracted_context)
    span_context = extracted_span.get_span_context()

    assert span_context.is_valid
    assert span_context.trace_id == VALID_TRACE_ID
    assert span_context.span_id == VALID_SPAN_ID


def test_inject_context_convenience_function(
    tracer_provider: TracerProvider,
) -> None:
    """Module-level inject_context delegates to propagator instance."""

    tracer = tracer_provider.get_tracer(__name__)

    with tracer.start_as_current_span("test-span") as span:
        current_context = trace.set_span_in_context(span)
        carrier: dict[str, str] = {}
        inject_context(carrier, context=current_context)

        assert set(carrier.keys()) == snapshot({"traceparent"})


def test_inject_extract_session_roundtrip(
    tracer_provider: TracerProvider,
) -> None:
    """Inject and extract session ID through carrier headers."""

    tracer = tracer_provider.get_tracer(__name__)

    with (
        session(id="test-session-123"),
        tracer.start_as_current_span("test-span") as span,
    ):
        current_context = trace.set_span_in_context(span)
        carrier: dict[str, str] = {}
        inject_context(carrier, context=current_context)

        assert set(carrier.keys()) == snapshot({"traceparent", SESSION_HEADER_NAME})
        assert carrier[SESSION_HEADER_NAME] == snapshot("test-session-123")

        extracted_session_id = extract_session_id(carrier)
        assert extracted_session_id == snapshot("test-session-123")


@pytest.mark.parametrize(
    "propagator_format",
    PROPAGATOR_FORMATS,
)
def test_extract_inject_roundtrip(
    tracer: trace.Tracer, propagator_format: PropagatorFormat
) -> None:
    """Verify extract and inject roundtrip for all propagator formats."""
    _configure_propagator(propagator_format)

    with tracer.start_as_current_span("original-span") as original_span:
        carrier: dict[str, str] = {}
        inject_context(carrier)

        extracted_context = extract_context(carrier)
        extracted_span = trace.get_current_span(extracted_context)

        original_span_context = original_span.get_span_context()
        extracted_span_context = extracted_span.get_span_context()

        assert original_span_context.trace_id == extracted_span_context.trace_id
        assert original_span_context.span_id == extracted_span_context.span_id


def test_distributed_trace_client_to_server_continuity(tracer: trace.Tracer) -> None:
    """Verify trace_id propagates from client to server in distributed system."""
    _configure_propagator("tracecontext")

    with tracer.start_as_current_span("client-span") as client_span:
        client_trace_id = client_span.get_span_context().trace_id

        carrier: dict[str, str] = {}
        inject_context(carrier)

        extracted_context = extract_context(carrier)

        with (
            propagated_context(parent=extracted_context),
            tracer.start_as_current_span("server-span") as server_span,
        ):
            server_trace_id = server_span.get_span_context().trace_id

            assert client_trace_id == server_trace_id


def test_multihop_trace_propagation(tracer: trace.Tracer) -> None:
    """Verify trace_id propagates across multiple service hops (A→B→C)."""
    _configure_propagator("tracecontext")

    with tracer.start_as_current_span("service-a-span") as service_a_span:
        service_a_trace_id = service_a_span.get_span_context().trace_id

        carrier_a_to_b: dict[str, str] = {}
        inject_context(carrier_a_to_b)

        context_b = extract_context(carrier_a_to_b)
        with (
            propagated_context(parent=context_b),
            tracer.start_as_current_span("service-b-span") as service_b_span,
        ):
            service_b_trace_id = service_b_span.get_span_context().trace_id

            carrier_b_to_c: dict[str, str] = {}
            inject_context(carrier_b_to_c)

            context_c = extract_context(carrier_b_to_c)
            with (
                propagated_context(parent=context_c),
                tracer.start_as_current_span("service-c-span") as service_c_span,
            ):
                service_c_trace_id = service_c_span.get_span_context().trace_id

                assert service_a_trace_id == service_b_trace_id
                assert service_b_trace_id == service_c_trace_id


def test_cross_format_propagation_b3_to_composite(tracer: trace.Tracer) -> None:
    """Verify trace propagates when client uses b3 and server uses composite."""
    _configure_propagator("b3")

    with tracer.start_as_current_span("client-span") as client_span:
        client_trace_id = client_span.get_span_context().trace_id

        carrier: dict[str, str] = {}
        inject_context(carrier)

        _configure_propagator("composite")

        extracted_context = extract_context(carrier)

        with (
            propagated_context(parent=extracted_context),
            tracer.start_as_current_span("server-span") as server_span,
        ):
            server_trace_id = server_span.get_span_context().trace_id

            assert client_trace_id == server_trace_id


def test_concurrent_requests_context_isolation(tracer: trace.Tracer) -> None:
    """Verify concurrent requests maintain isolated trace contexts."""
    _configure_propagator("tracecontext")

    trace_ids: list[int] = []
    lock = threading.Lock()

    def create_trace_and_propagate(index: int) -> None:
        with tracer.start_as_current_span(f"request-{index}") as span:
            trace_id = span.get_span_context().trace_id

            carrier: dict[str, str] = {}
            inject_context(carrier)

            extracted_context = extract_context(carrier)

            with (
                propagated_context(parent=extracted_context),
                tracer.start_as_current_span(f"handler-{index}") as handler_span,
            ):
                handler_trace_id = handler_span.get_span_context().trace_id

                with lock:
                    trace_ids.append(trace_id)

                assert trace_id == handler_trace_id

    threads = [
        threading.Thread(target=create_trace_and_propagate, args=(i,)) for i in range(5)
    ]

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()

    assert len(trace_ids) == 5
    assert len(set(trace_ids)) == 5


def test_malformed_header_creates_new_trace(tracer: trace.Tracer) -> None:
    """Verify server creates new trace when receiving malformed traceparent."""
    _configure_propagator("tracecontext")

    carrier: dict[str, str] = {"traceparent": "invalid-traceparent-format"}

    extracted_context = extract_context(carrier)

    with (
        propagated_context(parent=extracted_context),
        tracer.start_as_current_span("new-span") as new_span,
    ):
        new_trace_id = new_span.get_span_context().trace_id

        assert new_trace_id > 0
        assert new_span.is_recording()


def test_partial_context_extraction(tracer: trace.Tracer) -> None:
    """Verify graceful handling of partially valid trace context."""
    _configure_propagator("tracecontext")

    carrier: dict[str, str] = {
        "traceparent": "00-0af7651916cd43dd8448eb211c80319c-invalid-01"
    }

    extracted_context = extract_context(carrier)

    with (
        propagated_context(parent=extracted_context),
        tracer.start_as_current_span("span-from-partial") as span,
    ):
        trace_id = span.get_span_context().trace_id

        assert trace_id > 0
        assert span.is_recording()


def test_empty_carrier_extraction(tracer: trace.Tracer) -> None:
    """Verify extraction from empty carrier creates valid context for new trace."""
    _configure_propagator("tracecontext")

    carrier: dict[str, str] = {}

    extracted_context = extract_context(carrier)

    with (
        propagated_context(parent=extracted_context),
        tracer.start_as_current_span("span-from-empty") as span,
    ):
        trace_id = span.get_span_context().trace_id

        assert trace_id > 0
        assert span.is_recording()


def test_reset_propagator_clears_singleton(tracer_provider: TracerProvider) -> None:
    """Verify reset_propagator clears the singleton and allows fresh initialization."""
    os.environ[ENV_PROPAGATOR_FORMAT] = "tracecontext"

    first_propagator = get_propagator()

    reset_propagator()
    os.environ[ENV_PROPAGATOR_FORMAT] = "b3"

    second_propagator = get_propagator()

    tracer = tracer_provider.get_tracer(__name__)

    with tracer.start_as_current_span("test-span") as span:
        current_context = trace.set_span_in_context(span)

        carrier1: dict[str, str] = {}
        carrier2: dict[str, str] = {}

        first_propagator.inject_context(carrier1, context=current_context)
        second_propagator.inject_context(carrier2, context=current_context)

        assert set(carrier1.keys()) == snapshot({"traceparent"})
        assert set(carrier2.keys()) == snapshot({"b3"})



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 1,
                "output_tokens": 6658,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 4658,
                "provider_tool_usage": None,
                "raw": """\
cache_tokens_details=None cached_content_token_count=None candidates_token_count=2000 candidates_tokens_details=None prompt_token_count=1 prompt_tokens_details=[ModalityTokenCount(
  modality=<MediaModality.TEXT: 'TEXT'>,
  token_count=1
)] thoughts_token_count=4658 tool_use_prompt_token_count=None tool_use_prompt_tokens_details=None total_token_count=6659 traffic_type=None\
""",
                "total_tokens": 6659,
            },
            "messages": [
                UserMessage(content=[Text(text="")]),
                AssistantMessage(
                    content=[
                        Text(
                            text='''\
The problem asks us to calculate a "special sum" for a given integer `n`. The special sum is the sum of all integers `k` from 1 to `n` such that `k` is *not* a perfect square and `k` is *not* a perfect cube.

This problem can be efficiently solved using the Principle of Inclusion-Exclusion.

Let `S_total` be the sum of all integers from 1 to `n`.
Let `S_squares` be the sum of all perfect squares less than or equal to `n`.
Let `S_cubes` be the sum of all perfect cubes less than or equal to `n`.
Let `S_both` be the sum of all numbers that are *both* perfect squares *and* perfect cubes less than or equal to `n`.

The sum of numbers that are *either* perfect squares *or* perfect cubes (or both) is `S_squares + S_cubes - S_both`.
We want to find the sum of numbers that are *neither* perfect squares *nor* perfect cubes. This can be calculated as:
`Special Sum = S_total - (S_squares + S_cubes - S_both)`
`Special Sum = S_total - S_squares - S_cubes + S_both`

Let's break down each component:

1.  **`S_total` (Sum of integers from 1 to `n`):**
    This is a standard arithmetic series sum: `n * (n + 1) / 2`.

2.  **`S_squares` (Sum of perfect squares <= `n`):**
    These are `1^2, 2^2, ..., m^2` where `m` is the largest integer such that `m^2 <= n`. So, `m = floor(sqrt(n))`.
    The sum of the first `m` squares is given by the formula: `m * (m + 1) * (2*m + 1) / 6`.

3.  **`S_cubes` (Sum of perfect cubes <= `n`):**
    These are `1^3, 2^3, ..., p^3` where `p` is the largest integer such that `p^3 <= n`. So, `p = floor(cbrt(n))`.
    The sum of the first `p` cubes is given by the formula: `(p * (p + 1) / 2)^2`.

4.  **`S_both` (Sum of numbers that are both perfect squares and perfect cubes <= `n`):**
    A number `k` is both a perfect square (`a^2`) and a perfect cube (`b^3`) if and only if `k` is a perfect sixth power (`c^6`). This is because the exponents in its prime factorization must be multiples of both 2 and 3, meaning they must be multiples of `lcm(2, 3) = 6`.
    So, these are `1^6, 2^6, ..., q^6` where `q` is the largest integer such that `q^6 <= n`. So, `q = floor(n^(1/6))`.
    There isn't a simple closed-form formula for the sum of `i^6`. However, `q` will be relatively small even for very large `n`. For `n = 10^18`, `q = floor((10^18)^(1/6)) = floor(10^3) = 1000`. We can simply iterate from `i = 1` to `q` and sum `i^6`.

**Handling Large `n` and Floating Point Precision:**

*   `n` can be up to `10^18`, so Python's arbitrary-precision integers are necessary for sums.
*   Calculating `floor(n**(1/k))` (for `k=2, 3, 6`) requires care due to potential floating-point precision issues. A robust way is to estimate using `int(n**(1/k))` and then adjust the result with a small `while` loop to ensure it's the correct floor. Python 3.8+ offers `math.isqrt` for integer square roots, which is the most robust for `k=2`. For `k=3` and `k=6`, manual adjustment is suitable.

**Algorithm Steps:**

1.  Calculate `s_total = n * (n + 1) // 2`.
2.  Find `m = floor(sqrt(n))` using a robust method (e.g., `math.isqrt(n)` or the `get_kth_root_floor` helper function below).
3.  Calculate `s_squares = m * (m + 1) * (2 * m + 1) // 6`.
4.  Find `p = floor(cbrt(n))` using a robust method.
5.  Calculate `s_cubes = (p * (p + 1) // 2)**2`.
6.  Find `q = floor(n^(1/6))` using a robust method.
7.  Calculate `s_sixth_powers = sum(i**6 for i in range(1, q + 1))`.
8.  Return `s_total - s_squares - s_cubes + s_sixth_powers`.

**Robust `k`-th Root Floor Function:**

```python
def get_kth_root_floor(N, k):
    """
    Calculates floor(N^(1/k)) robustly for large N.
    """
    if N < 0:
        raise ValueError("N cannot be negative")
    if N == 0:
        return 0

    # Initial estimate using float power. This is usually close.
    root = int(N**(1/k))

    # Adjust the estimate to ensure it's the true floor.
    # Check if 'root' is too small
    while (root + 1)**k <= N:
        root += 1
    # Check if 'root' is too large
    while root**k > N:
        root -= 1
    return root
```

**Python Implementation:**

```python
import math

def get_kth_root_floor(N, k):
    """
    Calculates floor(N^(1/k)) robustly for large N.
    """
    if N < 0:
        raise ValueError("N cannot be negative")
    if N == 0:
        return 0
    \n\
    # For square root, math.isqrt is ideal in Python 3.8+
    if k == 2 and hasattr(math, 'isqrt'):
        return math.isqrt(N)

    # Initial estimate using float power. This is usually close.
    root = int(N**(1/k))

    # Adjust the estimate to ensure it's the true floor.
    # Check if 'root' is too small
    # Using 'root + 1' to avoid overflow issues if 'root**k' itself is already very large,
    # though for k=2,3,6 and N=10^18, this won't be an issue.
    while (root + 1)**k <= N:
        root += 1
    # Check if 'root' is too large
    while root**k > N:
        root -= 1
    return root

def special_sum(n: int) -> int:
    """
    Calculates the sum of k for all 1 <= k <= n such that k is not a perfect square
    and k is not a perfect cube.
    """
    if n < 1:
        return 0

    # 1. Total sum from 1 to n
    s_total = n * (n + 1) // 2

    # 2. Sum of perfect squares <= n
    m = get_kth_root_floor(n, 2)
    s_squares = m * (m + 1) * (2 * m + 1) // 6

    # 3. Sum of perfect cubes <= n
    p = get_kth_root_floor(n, 3)
    s_cubes = (p * (p + 1) // 2)**2

    # 4. Sum of numbers that are both perfect squares AND perfect cubes (i.e., perfect sixth powers) <= n
    q = get_kth_root_floor(n, 6)
    s_sixth_powers = 0
    for i in range(1, q + 1):
        s_sixth_powers += i**6
    \n\
    # Apply the Inclusion-Exclusion Principle
    result = s_total - s_squares - s_cubes + s_sixth_powers
    return result

```\
'''
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": None,
                                "function_response": None,
                                "inline_data": None,
                                "text": '''\
The problem asks us to calculate a "special sum" for a given integer `n`. The special sum is the sum of all integers `k` from 1 to `n` such that `k` is *not* a perfect square and `k` is *not* a perfect cube.

This problem can be efficiently solved using the Principle of Inclusion-Exclusion.

Let `S_total` be the sum of all integers from 1 to `n`.
Let `S_squares` be the sum of all perfect squares less than or equal to `n`.
Let `S_cubes` be the sum of all perfect cubes less than or equal to `n`.
Let `S_both` be the sum of all numbers that are *both* perfect squares *and* perfect cubes less than or equal to `n`.

The sum of numbers that are *either* perfect squares *or* perfect cubes (or both) is `S_squares + S_cubes - S_both`.
We want to find the sum of numbers that are *neither* perfect squares *nor* perfect cubes. This can be calculated as:
`Special Sum = S_total - (S_squares + S_cubes - S_both)`
`Special Sum = S_total - S_squares - S_cubes + S_both`

Let's break down each component:

1.  **`S_total` (Sum of integers from 1 to `n`):**
    This is a standard arithmetic series sum: `n * (n + 1) / 2`.

2.  **`S_squares` (Sum of perfect squares <= `n`):**
    These are `1^2, 2^2, ..., m^2` where `m` is the largest integer such that `m^2 <= n`. So, `m = floor(sqrt(n))`.
    The sum of the first `m` squares is given by the formula: `m * (m + 1) * (2*m + 1) / 6`.

3.  **`S_cubes` (Sum of perfect cubes <= `n`):**
    These are `1^3, 2^3, ..., p^3` where `p` is the largest integer such that `p^3 <= n`. So, `p = floor(cbrt(n))`.
    The sum of the first `p` cubes is given by the formula: `(p * (p + 1) / 2)^2`.

4.  **`S_both` (Sum of numbers that are both perfect squares and perfect cubes <= `n`):**
    A number `k` is both a perfect square (`a^2`) and a perfect cube (`b^3`) if and only if `k` is a perfect sixth power (`c^6`). This is because the exponents in its prime factorization must be multiples of both 2 and 3, meaning they must be multiples of `lcm(2, 3) = 6`.
    So, these are `1^6, 2^6, ..., q^6` where `q` is the largest integer such that `q^6 <= n`. So, `q = floor(n^(1/6))`.
    There isn't a simple closed-form formula for the sum of `i^6`. However, `q` will be relatively small even for very large `n`. For `n = 10^18`, `q = floor((10^18)^(1/6)) = floor(10^3) = 1000`. We can simply iterate from `i = 1` to `q` and sum `i^6`.

**Handling Large `n` and Floating Point Precision:**

*   `n` can be up to `10^18`, so Python's arbitrary-precision integers are necessary for sums.
*   Calculating `floor(n**(1/k))` (for `k=2, 3, 6`) requires care due to potential floating-point precision issues. A robust way is to estimate using `int(n**(1/k))` and then adjust the result with a small `while` loop to ensure it's the correct floor. Python 3.8+ offers `math.isqrt` for integer square roots, which is the most robust for `k=2`. For `k=3` and `k=6`, manual adjustment is suitable.

**Algorithm Steps:**

1.  Calculate `s_total = n * (n + 1) // 2`.
2.  Find `m = floor(sqrt(n))` using a robust method (e.g., `math.isqrt(n)` or the `get_kth_root_floor` helper function below).
3.  Calculate `s_squares = m * (m + 1) * (2 * m + 1) // 6`.
4.  Find `p = floor(cbrt(n))` using a robust method.
5.  Calculate `s_cubes = (p * (p + 1) // 2)**2`.
6.  Find `q = floor(n^(1/6))` using a robust method.
7.  Calculate `s_sixth_powers = sum(i**6 for i in range(1, q + 1))`.
8.  Return `s_total - s_squares - s_cubes + s_sixth_powers`.

**Robust `k`-th Root Floor Function:**

```python
def get_kth_root_floor(N, k):
    """
    Calculates floor(N^(1/k)) robustly for large N.
    """
    if N < 0:
        raise ValueError("N cannot be negative")
    if N == 0:
        return 0

    # Initial estimate using float power. This is usually close.
    root = int(N**(1/k))

    # Adjust the estimate to ensure it's the true floor.
    # Check if 'root' is too small
    while (root + 1)**k <= N:
        root += 1
    # Check if 'root' is too large
    while root**k > N:
        root -= 1
    return root
```

**Python Implementation:**

```python
import math

def get_kth_root_floor(N, k):
    """
    Calculates floor(N^(1/k)) robustly for large N.
    """
    if N < 0:
        raise ValueError("N cannot be negative")
    if N == 0:
        return 0
    \n\
    # For square root, math.isqrt is ideal in Python 3.8+
    if k == 2 and hasattr(math, 'isqrt'):
        return math.isqrt(N)

    # Initial estimate using float power. This is usually close.
    root = int(N**(1/k))

    # Adjust the estimate to ensure it's the true floor.
    # Check if 'root' is too small
    # Using 'root + 1' to avoid overflow issues if 'root**k' itself is already very large,
    # though for k=2,3,6 and N=10^18, this won't be an issue.
    while (root + 1)**k <= N:
        root += 1
    # Check if 'root' is too large
    while root**k > N:
        root -= 1
    return root

def special_sum(n: int) -> int:
    """
    Calculates the sum of k for all 1 <= k <= n such that k is not a perfect square
    and k is not a perfect cube.
    """
    if n < 1:
        return 0

    # 1. Total sum from 1 to n
    s_total = n * (n + 1) // 2

    # 2. Sum of perfect squares <= n
    m = get_kth_root_floor(n, 2)
    s_squares = m * (m + 1) * (2 * m + 1) // 6

    # 3. Sum of perfect cubes <= n
    p = get_kth_root_floor(n, 3)
    s_cubes = (p * (p + 1) // 2)**2

    # 4. Sum of numbers that are both perfect squares AND perfect cubes (i.e., perfect sixth powers) <= n
    q = get_kth_root_floor(n, 6)
    s_sixth_powers = 0
    for i in range(1, q + 1):
        s_sixth_powers += i**6
    \n\
    # Apply the Inclusion-Exclusion Principle
    result = s_total - s_squares - s_cubes + s_sixth_powers
    return result

```\
''',
                                "thought": None,
                                "thought_signature": None,
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)



"""End-to-end tests for LLM call with strict tool that has optional properties.

OpenAI strict mode requires ALL properties to be in the required array,
even if they have default values. This test verifies the fix works correctly.
"""

import pytest

from mirascope import llm
from tests.e2e.conftest import E2E_MODEL_IDS
from tests.utils import (
    Snapshot,
    snapshot_test,
)

# Include anthropic-beta/claude-sonnet-4-5 as it has strict support
E2E_MODEL_IDS = [*E2E_MODEL_IDS, "anthropic-beta/claude-sonnet-4-5"]


@llm.tool(strict=True)
def get_current_weather(location: str, unit: str = "fahrenheit") -> str:
    """Get the current weather in a given location.

    Args:
        location: The city and state, e.g. San Francisco, CA
        unit: The temperature unit to use (fahrenheit or celsius)
    """
    return f"The weather in {location} is 72°{unit[0].upper()}"


@pytest.mark.parametrize("model_id", E2E_MODEL_IDS)
@pytest.mark.vcr
def test_call_with_strict_tool_and_optional_properties(
    model_id: llm.ModelId,
    snapshot: Snapshot,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test a call with a strict tool that has optional properties.

    This validates that tools with default parameter values work correctly
    with OpenAI strict mode, which requires all properties to be required.
    """

    @llm.call(model_id, tools=[get_current_weather])
    def ask_weather() -> str:
        return "What's the weather like in San Francisco?"

    with snapshot_test(snapshot, caplog) as snap:
        response = ask_weather()
        snap.set_response(response)



"""Mock Provider for testing retry logic."""

from collections.abc import AsyncIterator, Iterator, Sequence
from typing import Any, ClassVar
from typing_extensions import Unpack

from mirascope import llm
from mirascope.llm import (
    AsyncContextResponse,
    AsyncContextStreamResponse,
    AsyncContextToolkit,
    AsyncResponse,
    AsyncStreamResponse,
    AsyncToolkit,
    Context,
    ContextResponse,
    ContextStreamResponse,
    ContextToolkit,
    DepsT,
    Format,
    FormattableT,
    OutputParser,
    Params,
    Response,
    StreamResponse,
    Toolkit,
)
from mirascope.llm.content import TextChunk, TextEndChunk, TextStartChunk
from mirascope.llm.providers import BaseProvider
from mirascope.llm.providers.base import ProviderErrorMap
from mirascope.llm.responses.base_stream_response import StreamResponseChunk


class MockProvider(BaseProvider[None]):
    """A mock Provider for testing that returns configurable responses.

    The provider can be configured with a sequence of exceptions to raise before
    eventually returning a successful response. This allows testing retry logic
    with predictable failure patterns.
    """

    # TODO Consider unifying with test provider.
    # cf #2101 (https://github.com/Mirascope/mirascope/issues/2120)

    id: ClassVar[llm.ProviderId] = "mock"
    default_scope: ClassVar[str | list[str]] = "mock/"
    error_map: ClassVar[ProviderErrorMap] = {}

    def __init__(self) -> None:
        """Initialize the MockProvider."""
        self.client = None
        self._exceptions: list[BaseException] = []
        self._stream_exceptions: list[BaseException] = []
        self._call_count = 0
        self._stream_count = 0
        self._stream_text: str | None = None  # Custom text for streaming (single)
        self._response_texts: list[
            str
        ] = []  # Queue of response texts for non-streaming
        self._stream_texts: list[str] = []  # Queue of stream texts

    @property
    def call_count(self) -> int:
        """The number of times _call() has been invoked."""
        return self._call_count

    @property
    def stream_count(self) -> int:
        """The number of times _stream() has been invoked."""
        return self._stream_count

    def set_exceptions(self, exceptions: Sequence[BaseException]) -> None:
        """Set exceptions to raise before returning a successful response.

        Args:
            exceptions: A list of exceptions to raise in order. Once exhausted,
                subsequent calls will return the mock response.
        """
        self._exceptions = list(exceptions)

    def set_stream_exceptions(self, exceptions: Sequence[BaseException]) -> None:
        """Set exceptions to raise during stream iteration.

        Args:
            exceptions: A list of exceptions to raise mid-stream. Each exception
                is raised once during iteration. Once exhausted, subsequent streams
                will complete successfully.
        """
        self._stream_exceptions = list(exceptions)

    def set_stream_text(self, text: str) -> None:
        """Set custom text to stream instead of default 'mock response'.

        Args:
            text: The text to stream in chunks.
        """
        self._stream_text = text

    def set_response_texts(self, texts: Sequence[str]) -> None:
        """Set a queue of response texts for non-streaming calls.

        Each call will use the next text in the queue. Once exhausted,
        subsequent calls will use the default 'mock response'.

        Args:
            texts: A list of response texts in order.
        """
        self._response_texts = list(texts)

    def set_stream_texts(self, texts: Sequence[str]) -> None:
        """Set a queue of stream texts for streaming calls.

        Each stream will use the next text in the queue. Once exhausted,
        subsequent streams will use the default 'mock response'.

        Args:
            texts: A list of stream texts in order.
        """
        self._stream_texts = list(texts)

    def _make_response(
        self,
        model_id: str,
        messages: Sequence[llm.messages.Message],
        format: Format[Any] | None = None,
    ) -> Response[Any]:
        """Create a minimal valid Response for testing."""
        # Use queued response text if available
        text = self._response_texts.pop(0) if self._response_texts else "mock response"
        return Response(
            raw={"mock": True},
            usage=None,
            provider_id="mock",
            model_id=model_id,
            provider_model_name="test-model",
            params={},
            tools=None,
            format=format,
            input_messages=list(messages),
            assistant_message=llm.messages.assistant(
                [llm.Text(text=text)],
                model_id=model_id,
                provider_id="mock",
            ),
            finish_reason=None,
        )

    def _make_async_response(
        self,
        model_id: str,
        messages: Sequence[llm.messages.Message],
        format: Format[Any] | None = None,
    ) -> AsyncResponse[Any]:
        """Create a minimal valid AsyncResponse for testing."""
        # Use queued response text if available
        text = self._response_texts.pop(0) if self._response_texts else "mock response"
        return AsyncResponse(
            raw={"mock": True},
            usage=None,
            provider_id="mock",
            model_id=model_id,
            provider_model_name="test-model",
            params={},
            tools=None,
            format=format,
            input_messages=list(messages),
            assistant_message=llm.messages.assistant(
                [llm.Text(text=text)],
                model_id=model_id,
                provider_id="mock",
            ),
            finish_reason=None,
        )

    def _make_context_response(
        self,
        model_id: str,
        messages: Sequence[llm.messages.Message],
        format: Format[Any] | None = None,
    ) -> ContextResponse[Any, Any]:
        """Create a minimal valid ContextResponse for testing."""
        # Use queued response text if available
        text = self._response_texts.pop(0) if self._response_texts else "mock response"
        return ContextResponse(
            raw={"mock": True},
            usage=None,
            provider_id="mock",
            model_id=model_id,
            provider_model_name="test-model",
            params={},
            tools=None,
            format=format,
            input_messages=list(messages),
            assistant_message=llm.messages.assistant(
                [llm.Text(text=text)],
                model_id=model_id,
                provider_id="mock",
            ),
            finish_reason=None,
        )

    def _make_async_context_response(
        self,
        model_id: str,
        messages: Sequence[llm.messages.Message],
        format: Format[Any] | None = None,
    ) -> AsyncContextResponse[Any, Any]:
        """Create a minimal valid AsyncContextResponse for testing."""
        # Use queued response text if available
        text = self._response_texts.pop(0) if self._response_texts else "mock response"
        return AsyncContextResponse(
            raw={"mock": True},
            usage=None,
            provider_id="mock",
            model_id=model_id,
            provider_model_name="test-model",
            params={},
            tools=None,
            format=format,
            input_messages=list(messages),
            assistant_message=llm.messages.assistant(
                [llm.Text(text=text)],
                model_id=model_id,
                provider_id="mock",
            ),
            finish_reason=None,
        )

    def _call(
        self,
        *,
        model_id: str,
        messages: Sequence[llm.messages.Message],
        toolkit: Toolkit,
        format: type[FormattableT]
        | Format[FormattableT]
        | OutputParser[FormattableT]
        | None = None,
        **params: Unpack[Params],
    ) -> Response[Any]:
        """Return the configured response or raise the next configured exception."""
        self._call_count += 1
        if self._exceptions:
            raise self._exceptions.pop(0)
        # Resolve format to Format object
        resolved_format: Format[Any] | None = None
        if format is not None:
            if isinstance(format, Format):
                resolved_format = format
            elif isinstance(format, OutputParser):
                resolved_format = llm.format(format, mode="parser")
            else:
                resolved_format = llm.format(format, mode="json")
        return self._make_response(model_id, messages, resolved_format)

    async def _call_async(
        self,
        *,
        model_id: str,
        messages: Sequence[llm.messages.Message],
        toolkit: AsyncToolkit,
        format: type[FormattableT]
        | Format[FormattableT]
        | OutputParser[FormattableT]
        | None = None,
        **params: Unpack[Params],
    ) -> AsyncResponse[Any]:
        """Return the configured response or raise the next configured exception."""
        self._call_count += 1
        if self._exceptions:
            raise self._exceptions.pop(0)
        # Resolve format to Format object
        resolved_format: Format[Any] | None = None
        if format is not None:
            if isinstance(format, Format):
                resolved_format = format
            elif isinstance(format, OutputParser):
                resolved_format = llm.format(format, mode="parser")
            else:
                resolved_format = llm.format(format, mode="json")
        return self._make_async_response(model_id, messages, resolved_format)

    def _context_call(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[llm.messages.Message],
        toolkit: ContextToolkit[DepsT],
        format: type[FormattableT]
        | Format[FormattableT]
        | OutputParser[FormattableT]
        | None = None,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Return the configured response or raise the next configured exception."""
        self._call_count += 1
        if self._exceptions:
            raise self._exceptions.pop(0)
        # Resolve format to Format object
        resolved_format: Format[Any] | None = None
        if format is not None:
            if isinstance(format, Format):
                resolved_format = format
            elif isinstance(format, OutputParser):
                resolved_format = llm.format(format, mode="parser")
            else:
                resolved_format = llm.format(format, mode="json")
        return self._make_context_response(model_id, messages, resolved_format)

    async def _context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[llm.messages.Message],
        toolkit: AsyncContextToolkit[DepsT],
        format: type[FormattableT]
        | Format[FormattableT]
        | OutputParser[FormattableT]
        | None = None,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Return the configured response or raise the next configured exception."""
        self._call_count += 1
        if self._exceptions:
            raise self._exceptions.pop(0)
        # Resolve format to Format object
        resolved_format: Format[Any] | None = None
        if format is not None:
            if isinstance(format, Format):
                resolved_format = format
            elif isinstance(format, OutputParser):
                resolved_format = llm.format(format, mode="parser")
            else:
                resolved_format = llm.format(format, mode="json")
        return self._make_async_context_response(model_id, messages, resolved_format)

    def _make_chunk_iterator(self) -> Iterator[StreamResponseChunk]:
        """Create a chunk iterator that yields mock text chunks.

        If stream exceptions are configured, raises the next one mid-stream.
        """
        yield TextStartChunk()
        # Determine text to stream: queued texts > single text > default
        if self._stream_texts:
            text = self._stream_texts.pop(0)
        elif self._stream_text is not None:
            text = self._stream_text
        else:
            text = "mock response"

        mid = len(text) // 2
        yield TextChunk(delta=text[:mid])
        # Raise exception mid-stream if configured
        if self._stream_exceptions:
            raise self._stream_exceptions.pop(0)
        yield TextChunk(delta=text[mid:])
        yield TextEndChunk()

    async def _make_async_chunk_iterator(self) -> AsyncIterator[StreamResponseChunk]:
        """Create an async chunk iterator that yields mock text chunks.

        If stream exceptions are configured, raises the next one mid-stream.
        """
        yield TextStartChunk()
        # Determine text to stream: queued texts > single text > default
        if self._stream_texts:
            text = self._stream_texts.pop(0)
        elif self._stream_text is not None:
            text = self._stream_text
        else:
            text = "mock response"

        mid = len(text) // 2
        yield TextChunk(delta=text[:mid])
        # Raise exception mid-stream if configured
        if self._stream_exceptions:
            raise self._stream_exceptions.pop(0)
        yield TextChunk(delta=text[mid:])
        yield TextEndChunk()

    def _stream(
        self,
        *,
        model_id: str,
        messages: Sequence[llm.messages.Message],
        toolkit: Toolkit,
        format: type[FormattableT]
        | Format[FormattableT]
        | OutputParser[FormattableT]
        | None = None,
        **params: Unpack[Params],
    ) -> StreamResponse[Any]:
        """Return a StreamResponse with configurable mid-stream exceptions."""
        self._stream_count += 1
        # Convert format to Format if it's a type
        resolved_format: Format[Any] | None = None
        if format is not None:
            if isinstance(format, Format):
                resolved_format = format
            elif isinstance(format, OutputParser):
                resolved_format = llm.format(format, mode="parser")
            else:
                resolved_format = llm.format(format, mode="json")
        return StreamResponse(
            provider_id="mock",
            model_id=model_id,
            provider_model_name="test-model",
            params={},
            tools=None,
            format=resolved_format,
            input_messages=list(messages),
            chunk_iterator=self._make_chunk_iterator(),
        )

    async def _stream_async(
        self,
        *,
        model_id: str,
        messages: Sequence[llm.messages.Message],
        toolkit: AsyncToolkit,
        format: type[FormattableT]
        | Format[FormattableT]
        | OutputParser[FormattableT]
        | None = None,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse[Any]:
        """Return an AsyncStreamResponse with configurable mid-stream exceptions."""
        self._stream_count += 1
        # Convert format to Format if it's a type
        resolved_format: Format[Any] | None = None
        if format is not None:
            if isinstance(format, Format):
                resolved_format = format
            elif isinstance(format, OutputParser):
                resolved_format = llm.format(format, mode="parser")
            else:
                resolved_format = llm.format(format, mode="json")
        return AsyncStreamResponse(
            provider_id="mock",
            model_id=model_id,
            provider_model_name="test-model",
            params={},
            tools=None,
            format=resolved_format,
            input_messages=list(messages),
            chunk_iterator=self._make_async_chunk_iterator(),
        )

    def _context_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[llm.messages.Message],
        toolkit: ContextToolkit[DepsT],
        format: type[FormattableT]
        | Format[FormattableT]
        | OutputParser[FormattableT]
        | None = None,
        **params: Unpack[Params],
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Return a ContextStreamResponse with configurable mid-stream exceptions."""
        self._stream_count += 1
        # Convert format to Format if it's a type
        resolved_format: Format[Any] | None = None
        if format is not None:
            if isinstance(format, Format):
                resolved_format = format
            elif isinstance(format, OutputParser):
                resolved_format = llm.format(format, mode="parser")
            else:
                resolved_format = llm.format(format, mode="json")
        return ContextStreamResponse(
            provider_id="mock",
            model_id=model_id,
            provider_model_name="test-model",
            params={},
            tools=None,
            format=resolved_format,
            input_messages=list(messages),
            chunk_iterator=self._make_chunk_iterator(),
        )

    async def _context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[llm.messages.Message],
        toolkit: AsyncContextToolkit[DepsT],
        format: type[FormattableT]
        | Format[FormattableT]
        | OutputParser[FormattableT]
        | None = None,
        **params: Unpack[Params],
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Return an AsyncContextStreamResponse with configurable mid-stream exceptions."""
        self._stream_count += 1
        # Convert format to Format if it's a type
        resolved_format: Format[Any] | None = None
        if format is not None:
            if isinstance(format, Format):
                resolved_format = format
            elif isinstance(format, OutputParser):
                resolved_format = llm.format(format, mode="parser")
            else:
                resolved_format = llm.format(format, mode="json")
        return AsyncContextStreamResponse(
            provider_id="mock",
            model_id=model_id,
            provider_model_name="test-model",
            params={},
            tools=None,
            format=resolved_format,
            input_messages=list(messages),
            chunk_iterator=self._make_async_chunk_iterator(),
        )

    def get_error_status(self, e: Exception) -> int | None:
        """Return None for mock provider."""
        return None



"""Interfaces for LLM providers."""

from ..._stubs import stub_module_if_missing

# Stub modules for missing optional dependencies BEFORE importing
# This must happen before any imports from these modules
# Note: We only stub top-level provider modules, not their submodules.
# The _StubModule will automatically handle nested attribute access.
stub_module_if_missing("mirascope.llm.providers.anthropic", "anthropic")
stub_module_if_missing("mirascope.llm.providers.google", "google")
stub_module_if_missing("mirascope.llm.providers.mlx", "mlx")
stub_module_if_missing("mirascope.llm.providers.openai", "openai")
stub_module_if_missing("mirascope.llm.providers.openrouter", "openai")
stub_module_if_missing("mirascope.llm.providers.together", "openai")
stub_module_if_missing("mirascope.llm.providers.ollama", "openai")

# Now imports work regardless of which packages are installed
# ruff: noqa: E402
from .anthropic import (
    AnthropicModelId,
    AnthropicProvider,
)
from .base import BaseProvider, Provider
from .google import GoogleModelId, GoogleProvider
from .mirascope import MirascopeProvider
from .mlx import MLXModelId, MLXProvider
from .model_id import ModelId
from .ollama import OllamaProvider
from .openai import (
    OpenAIModelId,
    OpenAIProvider,
)
from .openai.completions import BaseOpenAICompletionsProvider
from .openrouter import OpenRouterProvider
from .provider_id import KNOWN_PROVIDER_IDS, ProviderId
from .provider_registry import (
    get_provider_for_model,
    register_provider,
    reset_provider_registry,
)
from .together import TogetherProvider

__all__ = [
    "KNOWN_PROVIDER_IDS",
    "AnthropicModelId",
    "AnthropicProvider",
    "BaseOpenAICompletionsProvider",
    "BaseProvider",
    "GoogleModelId",
    "GoogleProvider",
    "MLXModelId",
    "MLXProvider",
    "MirascopeProvider",
    "ModelId",
    "OllamaProvider",
    "OpenAIModelId",
    "OpenAIProvider",
    "OpenRouterProvider",
    "Provider",
    "ProviderId",
    "TogetherProvider",
    "get_provider_for_model",
    "register_provider",
    "reset_provider_registry",
]



from mirascope import ops


@ops.trace
def handle_request(query: str) -> str:
    return f"Response to: {query}"


# Session with custom attributes
with ops.session(
    id="conversation-456",
    attributes={"user_tier": "premium", "channel": "web"},
) as session_ctx:
    print(f"Session ID: {session_ctx.id}")
    print(f"Attributes: {session_ctx.attributes}")

    trace = handle_request.wrapped("What's the weather?")
    print(trace.result)



import asyncio

from mirascope import llm

model = "openai/gpt-5-mini"


@llm.prompt
async def chef_selector(ingredient: str):
    return (
        f"Identify a chef known for cooking with {ingredient}. Return only their name."
    )


@llm.prompt(format=list[str])
async def ingredients_identifier(ingredient: str):
    return f"List 5 ingredients that complement {ingredient}."


@llm.prompt
async def recommend(chef: str, ingredients: list[str]):
    return f"As chef {chef}, recommend a recipe using: {ingredients}"


async def recipe_recommender(ingredient: str) -> str:
    chef_response, ingredients_response = await asyncio.gather(
        chef_selector(model, ingredient),
        ingredients_identifier(model, ingredient),
    )

    response = await recommend(
        model, chef_response.text(), ingredients_response.parse()
    )
    return response.text()


async def main():
    recipe = await recipe_recommender("apples")
    print(recipe)


asyncio.run(main())



"""OpenAI Responses message encoding and request preparation."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, TypedDict, cast

from openai import Omit
from openai.types.responses import (
    FunctionToolParam,
    ResponseFormatTextJSONSchemaConfigParam,
    ResponseFunctionToolCallParam,
    ResponseInputContentParam,
    ResponseInputItemParam,
    ResponseInputParam,
    ResponseInputTextParam,
    ResponseTextConfigParam,
    ToolChoiceAllowedParam,
    ToolChoiceFunctionParam,
    ToolParam,
    WebSearchToolParam,
    response_create_params,
)
from openai.types.responses.easy_input_message_param import EasyInputMessageParam
from openai.types.responses.response_input_image_param import ResponseInputImageParam
from openai.types.responses.response_input_param import (
    FunctionCallOutput,
    Message as ResponseInputMessageParam,
)
from openai.types.shared_params import Reasoning, ReasoningEffort
from openai.types.shared_params.response_format_json_object import (
    ResponseFormatJSONObject,
)
from openai.types.shared_params.responses_model import ResponsesModel

from .....exceptions import FeatureNotSupportedError
from .....formatting import (
    Format,
    FormatSpec,
    FormattableT,
    resolve_format,
)
from .....messages import AssistantMessage, Message, UserMessage
from .....tools import (
    FORMAT_TOOL_NAME,
    AnyToolSchema,
    BaseToolkit,
    ProviderTool,
    WebSearchTool,
)
from ....base import _utils as _base_utils
from ...model_id import OpenAIModelId, model_name
from ...model_info import (
    MODELS_WITHOUT_JSON_OBJECT_SUPPORT,
    MODELS_WITHOUT_JSON_SCHEMA_SUPPORT,
    NON_REASONING_MODELS,
)

if TYPE_CHECKING:
    from .....models import Params, ThinkingLevel

# Thinking level to a float multiplier % of max tokens
THINKING_LEVEL_TO_EFFORT: dict[ThinkingLevel, ReasoningEffort] = {
    "default": "medium",
    "none": "none",
    "minimal": "minimal",
    "low": "low",
    "medium": "medium",
    "high": "high",
    "max": "xhigh",
}


class ResponseCreateKwargs(TypedDict, total=False):
    """Kwargs to the OpenAI `client.responses.create` method."""

    model: ResponsesModel
    input: str | ResponseInputParam
    instructions: str
    temperature: float
    max_output_tokens: int
    top_p: float
    tools: list[ToolParam] | Omit
    tool_choice: response_create_params.ToolChoice | Omit
    text: ResponseTextConfigParam
    reasoning: Reasoning | Omit


def _encode_user_message(
    message: UserMessage,
) -> ResponseInputParam:
    if len(message.content) == 1 and (first := message.content[0]).type == "text":
        return [EasyInputMessageParam(content=first.text, role="user")]

    current_content: list[ResponseInputContentParam] = []
    result: ResponseInputParam = []

    def flush_message_content() -> None:
        nonlocal current_content
        if current_content:
            result.append(
                ResponseInputMessageParam(
                    content=current_content, role="user", type="message"
                )
            )
            current_content = []

    for part in message.content:
        if part.type == "text":
            current_content.append(
                ResponseInputTextParam(text=part.text, type="input_text")
            )
        elif part.type == "image":
            image_url = (
                part.source.url
                if part.source.type == "url_image_source"
                else f"data:{part.source.mime_type};base64,{part.source.data}"
            )

            current_content.append(
                ResponseInputImageParam(
                    image_url=image_url, detail="auto", type="input_image"
                )
            )
        elif part.type == "tool_output":
            flush_message_content()
            result.append(
                FunctionCallOutput(
                    call_id=part.id,
                    output=str(part.result),
                    type="function_call_output",
                )
            )
        elif part.type == "audio":
            raise FeatureNotSupportedError(
                "audio input",
                "openai",
                message='provider "openai" does not support audio inputs when using :responses api. Try appending :completions to your model instead.',
            )
        else:
            raise NotImplementedError(
                f"Unsupported user content part type: {part.type}"
            )
    flush_message_content()

    return result


def _encode_assistant_message(
    message: AssistantMessage, encode_thoughts: bool
) -> ResponseInputParam:
    result: ResponseInputParam = []

    # Note: OpenAI does not provide any way to encode multiplie pieces of assistant-generated
    # text as adjacent content within the same Message, except as part of
    # ResponseOutputMessageParam which requires OpenAI-provided `id` and `status` on the message,
    # and `annotations` and `logprobs` on the output text.
    # Rather than generating a fake or nonexistent fields and triggering potentially undefined
    # server-side behavior, we use `EasyInputMessageParam` for assistant generated text,
    # with the caveat that assistant messages containing multiple text parts will be encoded
    # as though they are separate messages.
    # (It would seem as though the `Message` class in `response_input_param.py` would be suitable,
    # especially as it supports the "assistant" role; however attempting to use it triggers a server
    # error when text of type input_text is passed as part of an assistant message.)
    for part in message.content:
        if part.type == "text":
            result.append(EasyInputMessageParam(content=part.text, role="assistant"))
        elif part.type == "thought":
            if encode_thoughts:
                result.append(
                    EasyInputMessageParam(
                        content="**Thinking:** " + part.thought, role="assistant"
                    )
                )
        elif part.type == "tool_call":
            result.append(
                ResponseFunctionToolCallParam(
                    call_id=part.id,
                    name=part.name,
                    arguments=part.args,
                    type="function_call",
                )
            )
        else:
            raise NotImplementedError(
                f"Unsupported assistant content part type: {part.type}"
            )

    return result


def _encode_message(
    message: Message, model_id: OpenAIModelId, encode_thoughts: bool
) -> ResponseInputParam:
    """Convert a Mirascope Message to OpenAI Responses input items.

    Returns a list because tool calls and tool outputs become separate input items
    in the Responses API, not part of message content.
    """

    if message.role == "system":
        # Responses API allows multiple "developer" messages, so rather than using the
        # instructions field, we convert system messages as we find them.
        # Unlike other LLM APIs, the system message does not need to be the first message.
        return [EasyInputMessageParam(role="developer", content=message.content.text)]

    if (
        message.role == "assistant"
        and message.provider_id in ("openai", "openai:responses")
        and message.provider_model_name
        == model_name(model_id=model_id, api_mode="responses")
        and message.raw_message
        and not encode_thoughts
    ):
        return cast(ResponseInputParam, message.raw_message)

    if message.role == "assistant":
        return _encode_assistant_message(message, encode_thoughts)
    else:
        return _encode_user_message(message)


def _convert_tool_to_tool_param(
    tool: AnyToolSchema | ProviderTool,
) -> ToolParam:
    """Convert a Mirascope ToolSchema to OpenAI Responses ToolParam."""
    if isinstance(tool, WebSearchTool):
        return WebSearchToolParam(type="web_search")
    if isinstance(tool, ProviderTool):
        raise FeatureNotSupportedError(
            f"Provider tool {tool.name}", provider_id="openai:responses"
        )
    schema_dict = tool.parameters.model_dump(by_alias=True, exclude_none=True)
    schema_dict["type"] = "object"
    _base_utils.ensure_additional_properties_false(schema_dict)
    strict = True if tool.strict is None else tool.strict
    if strict:
        _base_utils.ensure_all_properties_required(schema_dict)

    return FunctionToolParam(
        type="function",
        name=tool.name,
        description=tool.description,
        parameters=schema_dict,
        strict=strict,
    )


def _create_strict_response_format(
    format: Format[FormattableT],
) -> ResponseFormatTextJSONSchemaConfigParam:
    """Create OpenAI Responses strict response format from a Mirascope Format.

    Args:
        format: The `Format` instance containing schema and metadata

    Returns:
        ResponseFormatTextJSONSchemaConfigParam for strict structured outputs
    """
    schema = format.schema.copy()
    _base_utils.ensure_additional_properties_false(schema)

    response_format: ResponseFormatTextJSONSchemaConfigParam = {
        "type": "json_schema",
        "name": format.name,
        "schema": schema,
    }
    if format.description:
        response_format["description"] = format.description
    response_format["strict"] = True

    return response_format


def _compute_reasoning(
    level: ThinkingLevel,
    include_thoughts: bool,
) -> Reasoning:
    """Compute the OpenAI `Reasoning` config based on ThinkingConfig.

    Args:
        level: The thinking level
        include_thoughts: Whether to include summary (True/False for auto)

    Returns:
        OpenAI Reasoning configuration
    """
    reasoning: Reasoning = {"effort": THINKING_LEVEL_TO_EFFORT.get(level) or "medium"}

    if include_thoughts:
        reasoning["summary"] = "auto"

    return reasoning


def encode_request(
    *,
    model_id: OpenAIModelId,
    messages: Sequence[Message],
    tools: BaseToolkit[AnyToolSchema],
    format: FormatSpec[FormattableT] | None,
    params: Params,
) -> tuple[Sequence[Message], Format[FormattableT] | None, ResponseCreateKwargs]:
    """Prepares a request for the `OpenAI.responses.create` method."""
    if model_id.endswith(":completions"):
        raise FeatureNotSupportedError(
            feature="completions API",
            provider_id="openai:responses",
            model_id=model_id,
            message=f"Cannot use completions model with responses client: {model_id}",
        )

    base_model_name = model_name(model_id, None)

    kwargs: ResponseCreateKwargs = ResponseCreateKwargs(
        {
            "model": base_model_name,
        }
    )
    encode_thoughts = False

    with _base_utils.ensure_all_params_accessed(
        params=params,
        provider_id="openai",
        unsupported_params=["top_k", "seed", "stop_sequences"],
    ) as param_accessor:
        if param_accessor.temperature is not None:
            kwargs["temperature"] = param_accessor.temperature
        if param_accessor.max_tokens is not None:
            kwargs["max_output_tokens"] = param_accessor.max_tokens
        if param_accessor.top_p is not None:
            kwargs["top_p"] = param_accessor.top_p
        if param_accessor.thinking is not None:
            thinking_config = param_accessor.thinking
            if base_model_name in NON_REASONING_MODELS:
                param_accessor.emit_warning_for_unused_param(
                    "thinking", thinking_config, "openai", model_id
                )
            else:
                # Assume model supports reasoning unless explicitly listed as non-reasoning
                # This ensures new reasoning models work immediately without code updates
                level = thinking_config.get("level")
                include_thoughts = thinking_config.get("include_thoughts", False)
                kwargs["reasoning"] = _compute_reasoning(level, include_thoughts)

            # Handle encode_thoughts_as_text from ThinkingConfig
            if thinking_config.get("encode_thoughts_as_text"):
                encode_thoughts = True

    openai_tools = [_convert_tool_to_tool_param(tool) for tool in tools.tools]

    model_supports_strict = model_id not in MODELS_WITHOUT_JSON_SCHEMA_SUPPORT
    default_mode = "strict" if model_supports_strict else "tool"

    format = resolve_format(format, default_mode=default_mode)
    if format is not None:
        if format.mode == "strict":
            kwargs["text"] = {"format": _create_strict_response_format(format)}
        elif format.mode == "tool":
            format_tool_schema = format.create_tool_schema()
            openai_tools.append(_convert_tool_to_tool_param(format_tool_schema))
            if tools.tools:
                # Only include function tools in tool_choice (filter out web search, etc.)
                function_tools = cast(
                    list[FunctionToolParam],
                    [t for t in openai_tools if t.get("type") == "function"],
                )
                kwargs["tool_choice"] = ToolChoiceAllowedParam(
                    type="allowed_tools",
                    mode="required",
                    tools=[
                        {"type": "function", "name": tool["name"]}
                        for tool in function_tools
                    ],
                )
            else:
                kwargs["tool_choice"] = ToolChoiceFunctionParam(
                    type="function",
                    name=FORMAT_TOOL_NAME,
                )
        elif (
            format.mode == "json" and model_id not in MODELS_WITHOUT_JSON_OBJECT_SUPPORT
        ):
            kwargs["text"] = {"format": ResponseFormatJSONObject(type="json_object")}

        if format.formatting_instructions:
            messages = _base_utils.add_system_instructions(
                messages, format.formatting_instructions
            )

    encoded_messages: list[ResponseInputItemParam] = []
    for message in messages:
        encoded_messages.extend(_encode_message(message, model_id, encode_thoughts))
    kwargs["input"] = encoded_messages

    if openai_tools:
        kwargs["tools"] = openai_tools

    return messages, format, kwargs



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:responses",
            "provider_model_name": "gpt-4o:responses",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 7,
                "output_tokens": 26,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "ResponseUsage(input_tokens=7, input_tokens_details=InputTokensDetails(cached_tokens=0), output_tokens=26, output_tokens_details=OutputTokensDetails(reasoning_tokens=0), total_tokens=33)",
                "total_tokens": 33,
            },
            "messages": [
                UserMessage(content=[Text(text="")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="It looks like you're trying to share an image or details about a scene or object. How can I assist you with it?"
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "id": "msg_02aecbd82c76cb54006931df309ef08197a1a028224b73014e",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": "It looks like you're trying to share an image or details about a scene or object. How can I assist you with it?",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        }
                    ],
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    ToolCall,
    ToolOutput,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:responses",
            "provider_model_name": "gpt-4o:responses",
            "params": {"thinking": {"level": "none"}},
            "finish_reason": None,
            "usage": {
                "input_tokens": 84,
                "output_tokens": 18,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "ResponseUsage(input_tokens=84, input_tokens_details=InputTokensDetails(cached_tokens=0), output_tokens=18, output_tokens_details=OutputTokensDetails(reasoning_tokens=0), total_tokens=102)",
                "total_tokens": 102,
            },
            "messages": [
                UserMessage(content=[Text(text="What is the 100th fibonacci number?")]),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="google_unknown_tool_id",
                            name="compute_fib",
                            args='{"n": 100}',
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": {
                                    "id": None,
                                    "args": {"n": 100},
                                    "name": "compute_fib",
                                    "partial_args": None,
                                    "will_continue": None,
                                },
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b'\n\xe2\x01\x01r\xc8\xda|\xb3\xc4\xd3\xe0\xa7\xc0{I\x80\xd2\r\x8e|4\xfa1W\\\xd7\xd4\xcd\xfa|0]T\x8e\x9bq\xc3\xd4\xdc\n\x8c\x94\xc6t&\xf6\xf8a\x19\xe6R\xab\x99\xab\x9c\xf6\x17\x1b\xc7\xc3bL\x83\x80\x9bKQF\x19\x93\xef\xe1&\x94\xa7\x8c\x15\x827~\xc0\x17j\xc4zz\xca?\x08\x17\x8c\xa6\x08F>\xfa\x1e\x14iO\x0ee\x86\xbfv\xab\\\xcaj\x88\xc3\x9d9k\xe0\xad!Z5!\xac{e\xbe\x16\x1d\xf8+V\r\xa7B\xd5\xf6\x1e\xd5\xed\xad\xc8,\xc9X\xe8t\xb0F\x95\n\x15\x89/j7UU.\x07\xa1\x8b"u\xa47\xc9\xc6r:\xb0\x17\xda\x1cD\xa2\xc2\xbe\x0e \xa1\xbc]\x19o\x00\x90\x00\x0f\x10\x08\xe7"\x81\xde\xd6\x92\x16R\x81$\xd2t\xbf\xef\x074\xe9\x17\x03!\xed\xce\xa1)\xb8&2\xd1\xe8K\x18%\xe0]\x06\xc3O\xadz\x1c',
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="google_unknown_tool_id",
                            name="compute_fib",
                            result="218922995834555169026",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="The 100th Fibonacci number is 218922995834555169026."
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "id": "msg_06239e8744215ddf0069689ad7c6f88193853e05a8db2a352c",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": "The 100th Fibonacci number is 218922995834555169026.",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        }
                    ],
                ),
            ],
            "format": None,
            "tools": [
                {
                    "name": "compute_fib",
                    "description": "Compute the nth Fibonacci number (1-indexed).",
                    "parameters": """\
{
  "properties": {
    "n": {
      "title": "N",
      "type": "integer"
    }
  },
  "required": [
    "n"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": None,
                }
            ],
        }
    }
)



# This file was auto-generated by Fern from our API Definition.

import typing

OrganizationsPreviewSubscriptionChangeResponseValidationErrorsItemResource = (
    typing.Union[typing.Literal["seats", "projects"], typing.Any]
)



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "mlx",
            "model_id": "mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
            "provider_model_name": "mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 8,
                "output_tokens": 351,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "MLXCassetteResponse(text='', token=151645, prompt_tokens=8, generation_tokens=351, finish_reason='stop')",
                "total_tokens": 359,
            },
            "messages": [
                UserMessage(content=[Text(text="")]),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
<think>
Okay, the user is asking about the difference between the two main types of the same type. Let me think. The first type is the "main type," and the second is the "sub-type." I need to explain each in simple terms.

For the main type, it's the most general category. For example, if I'm talking about a book, the main type would be "book." The sub-type would be specific, like "fictional books." That makes sense. \n\

I should also mention that sub-types can be more specific and narrower. For instance, if the main type is "books," sub-types could be "novels" or "non-fiction." \n\

Wait, maybe I should check if there's a standard way to categorize these. I think in programming, the main type is the base type, and sub-types are derived from it. So the difference is about the base and derived types. \n\

I need to make sure the explanation is clear and not too technical. Let me structure it step by step: define main type, sub-type, examples, and why they differ. That should cover the user's question.
</think>

The two main types of the same type are:  \n\
1. **Main Type**: This is the general, most basic category. For example, if you're talking about books, the main type is "book."  \n\
2. **Sub-Type**: This is a specific, narrower category derived from the main type. For example, if the main type is "book," a sub-type could be "fictional books."  \n\

Sub-types are more specific and narrower, allowing for more precise categorization. They are often used to distinguish between different forms of a broader category.\
"""
                        )
                    ],
                    provider_id="mlx",
                    model_id="mlx-community/Qwen3-0.6B-4bit-DWQ-053125",
                    provider_model_name=None,
                    raw_message=None,
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    Text,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 17,
                "output_tokens": 15,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "Usage(cache_creation=CacheCreation(ephemeral_1h_input_tokens=0, ephemeral_5m_input_tokens=0), cache_creation_input_tokens=0, cache_read_input_tokens=0, input_tokens=17, output_tokens=15, server_tool_use=None, service_tier='standard')",
                "total_tokens": 32,
            },
            "messages": [
                UserMessage(content=[Text(text="What is 4200 + 42?")]),
                AssistantMessage(
                    content=[Text(text="4200 + 42 = 4242")],
                    provider_id="anthropic",
                    model_id="anthropic/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "citations": None,
                                "text": "4200 + 42 = 4242",
                                "type": "text",
                            }
                        ],
                    },
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 17,
                "output_tokens": 15,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "Usage(cache_creation=CacheCreation(ephemeral_1h_input_tokens=0, ephemeral_5m_input_tokens=0), cache_creation_input_tokens=0, cache_read_input_tokens=0, input_tokens=17, output_tokens=15, server_tool_use=None, service_tier='standard')",
                "total_tokens": 32,
            },
            "messages": [
                UserMessage(content=[Text(text="What is 4200 + 42?")]),
                AssistantMessage(
                    content=[Text(text="4200 + 42 = 4242")],
                    provider_id="anthropic",
                    model_id="anthropic/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [
                            {
                                "citations": None,
                                "text": "4200 + 42 = 4242",
                                "type": "text",
                            }
                        ],
                    },
                ),
            ],
            "format": None,
            "tools": [],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is 4200 + 42?")]),
                AssistantMessage(
                    content=[Text(text="4200 + 42 = 4242")],
                    provider_id="anthropic",
                    model_id="anthropic/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [{"type": "text", "text": "4200 + 42 = 4242"}],
                    },
                ),
            ],
            "format": None,
            "tools": [],
            "usage": {
                "input_tokens": 17,
                "output_tokens": 15,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 32,
            },
            "n_chunks": 4,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "anthropic",
            "model_id": "anthropic/claude-sonnet-4-0",
            "provider_model_name": "claude-sonnet-4-0",
            "finish_reason": None,
            "messages": [
                UserMessage(content=[Text(text="What is 4200 + 42?")]),
                AssistantMessage(
                    content=[Text(text="4200 + 42 = 4242")],
                    provider_id="anthropic",
                    model_id="anthropic/claude-sonnet-4-0",
                    provider_model_name="claude-sonnet-4-0",
                    raw_message={
                        "role": "assistant",
                        "content": [{"type": "text", "text": "4200 + 42 = 4242"}],
                    },
                ),
            ],
            "format": None,
            "tools": [],
            "usage": {
                "input_tokens": 17,
                "output_tokens": 15,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "provider_tool_usage": None,
                "raw": "None",
                "total_tokens": 32,
            },
            "n_chunks": 4,
        }
    }
)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
import typing_extensions
from ...core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from ...core.serialization import FieldMetadata
from ...types.date import Date
from .project_memberships_update_response_role import (
    ProjectMembershipsUpdateResponseRole,
)


class ProjectMembershipsUpdateResponse(UniversalBaseModel):
    member_id: typing_extensions.Annotated[str, FieldMetadata(alias="memberId")]
    organization_id: typing_extensions.Annotated[
        str, FieldMetadata(alias="organizationId")
    ]
    project_id: typing_extensions.Annotated[str, FieldMetadata(alias="projectId")]
    role: ProjectMembershipsUpdateResponseRole
    created_at: typing_extensions.Annotated[
        typing.Optional[Date], FieldMetadata(alias="createdAt")
    ] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
            extra="allow", frozen=True
        )  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from mirascope import ops

# Configure Mirascope Cloud (uses MIRASCOPE_API_KEY env var)
ops.configure()

# Or pass the API key directly:
# ops.configure(api_key="your-api-key")

# Now all traced operations will be sent to Mirascope Cloud



"""Concrete Prompt classes for generating messages with tools and formatting."""

from collections.abc import Callable, Sequence
from typing import Any, Generic, TypeVar, overload

from ..._utils import copy_function_metadata
from ..context import Context, DepsT
from ..formatting import FormatSpec, FormattableT
from ..messages import Message, promote_to_messages
from ..models import Model
from ..providers import ModelId
from ..responses import (
    AsyncContextResponse,
    AsyncContextStreamResponse,
    AsyncResponse,
    AsyncStreamResponse,
    ContextResponse,
    ContextStreamResponse,
    Response,
    StreamResponse,
)
from ..tools import AsyncContextToolkit, AsyncToolkit, ContextToolkit, Toolkit
from ..types import P
from .protocols import (
    AsyncContextMessageTemplate,
    AsyncMessageTemplate,
    ContextMessageTemplate,
    MessageTemplate,
)

FunctionT = TypeVar("FunctionT", bound=Callable[..., Any])


class BasePrompt(Generic[FunctionT]):
    """Base class for all Prompt types with shared metadata functionality."""

    fn: FunctionT
    """The underlying prompt function that generates message content."""

    __name__: str = ""
    """The name of the underlying function (preserved for decorator stacking)."""

    def __init__(self, *, fn: FunctionT) -> None:
        self.fn = fn
        copy_function_metadata(self, self.fn)


class Prompt(BasePrompt[MessageTemplate[P]], Generic[P, FormattableT]):
    """A prompt that can be called with a model to generate a response.

    Created by decorating a `MessageTemplate` with `llm.prompt`. The decorated
    function becomes callable with a `Model` to generate LLM responses.

    A `Prompt` is essentially: `MessageTemplate` + tools + format.
    It can be invoked with a model: `prompt(model, *args, **kwargs)`.
    """

    toolkit: Toolkit
    """The toolkit containing this prompt's tools."""

    format: FormatSpec[FormattableT] | None
    """The response format for the generated response."""

    def __init__(
        self,
        *,
        fn: MessageTemplate[P],
        toolkit: Toolkit,
        format: FormatSpec[FormattableT] | None,
    ) -> None:
        self.toolkit = toolkit
        self.format = format
        super().__init__(fn=fn)

    def messages(self, *args: P.args, **kwargs: P.kwargs) -> Sequence[Message]:
        """Return the `Messages` from invoking this prompt."""
        return promote_to_messages(self.fn(*args, **kwargs))

    @overload
    def __call__(
        self: "Prompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Response: ...

    @overload
    def __call__(
        self: "Prompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Response[FormattableT]: ...

    def __call__(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> Response | Response[FormattableT]:
        """Generates a response using the provided model."""
        return self.call(model, *args, **kwargs)

    @overload
    def call(
        self: "Prompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Response: ...

    @overload
    def call(
        self: "Prompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Response[FormattableT]: ...

    def call(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> Response | Response[FormattableT]:
        """Generates a response using the provided model."""
        if isinstance(model, str):
            model = Model(model)
        messages = self.messages(*args, **kwargs)
        return model.call(messages, tools=self.toolkit, format=self.format)

    @overload
    def stream(
        self: "Prompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> StreamResponse: ...

    @overload
    def stream(
        self: "Prompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> StreamResponse[FormattableT]: ...

    def stream(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Generates a streaming response using the provided model."""
        if isinstance(model, str):
            model = Model(model)
        messages = self.messages(*args, **kwargs)
        return model.stream(messages, tools=self.toolkit, format=self.format)


class AsyncPrompt(BasePrompt[AsyncMessageTemplate[P]], Generic[P, FormattableT]):
    """An async prompt that can be called with a model to generate a response.

    Created by decorating an async `MessageTemplate` with `llm.prompt`. The decorated
    async function becomes callable with a `Model` to generate LLM responses asynchronously.

    An `AsyncPrompt` is essentially: async `MessageTemplate` + tools + format.
    It can be invoked with a model: `await prompt(model, *args, **kwargs)`.
    """

    toolkit: AsyncToolkit
    """The toolkit containing this prompt's async tools."""

    format: FormatSpec[FormattableT] | None
    """The response format for the generated response."""

    def __init__(
        self,
        *,
        fn: AsyncMessageTemplate[P],
        toolkit: AsyncToolkit,
        format: FormatSpec[FormattableT] | None,
    ) -> None:
        self.toolkit = toolkit
        self.format = format
        super().__init__(fn=fn)

    async def messages(self, *args: P.args, **kwargs: P.kwargs) -> Sequence[Message]:
        """Return the `Messages` from invoking this prompt."""
        return promote_to_messages(await self.fn(*args, **kwargs))

    @overload
    async def __call__(
        self: "AsyncPrompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncResponse: ...

    @overload
    async def __call__(
        self: "AsyncPrompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncResponse[FormattableT]: ...

    async def __call__(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Generates a response using the provided model asynchronously."""
        return await self.call(model, *args, **kwargs)

    @overload
    async def call(
        self: "AsyncPrompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncResponse: ...

    @overload
    async def call(
        self: "AsyncPrompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncResponse[FormattableT]: ...

    async def call(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Generates a response using the provided model asynchronously."""
        if isinstance(model, str):
            model = Model(model)
        messages = await self.messages(*args, **kwargs)
        return await model.call_async(messages, tools=self.toolkit, format=self.format)

    @overload
    async def stream(
        self: "AsyncPrompt[P, None]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncStreamResponse: ...

    @overload
    async def stream(
        self: "AsyncPrompt[P, FormattableT]",
        model: Model | ModelId,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncStreamResponse[FormattableT]: ...

    async def stream(
        self, model: Model | ModelId, *args: P.args, **kwargs: P.kwargs
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Generates a streaming response using the provided model asynchronously."""
        if isinstance(model, str):
            model = Model(model)
        messages = await self.messages(*args, **kwargs)
        return await model.stream_async(
            messages, tools=self.toolkit, format=self.format
        )


class ContextPrompt(
    BasePrompt[ContextMessageTemplate[P, DepsT]], Generic[P, DepsT, FormattableT]
):
    """A context-aware prompt that can be called with a model to generate a response.

    Created by decorating a `ContextMessageTemplate` with `llm.prompt`. The decorated
    function (with first parameter `'ctx'` of type `Context[DepsT]`) becomes callable
    with a `Model` to generate LLM responses with context dependencies.

    A `ContextPrompt` is essentially: `ContextMessageTemplate` + tools + format.
    It can be invoked with a model: `prompt(model, ctx, *args, **kwargs)`.
    """

    toolkit: ContextToolkit[DepsT]
    """The toolkit containing this prompt's context-aware tools."""

    format: FormatSpec[FormattableT] | None
    """The response format for the generated response."""

    def __init__(
        self,
        *,
        fn: ContextMessageTemplate[P, DepsT],
        toolkit: ContextToolkit[DepsT],
        format: FormatSpec[FormattableT] | None,
    ) -> None:
        self.toolkit = toolkit
        self.format = format
        super().__init__(fn=fn)

    def messages(
        self, ctx: Context[DepsT], *args: P.args, **kwargs: P.kwargs
    ) -> Sequence[Message]:
        """Return the `Messages` from invoking this prompt."""
        return promote_to_messages(self.fn(ctx, *args, **kwargs))

    @overload
    def __call__(
        self: "ContextPrompt[P, DepsT, None]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ContextResponse[DepsT, None]: ...

    @overload
    def __call__(
        self: "ContextPrompt[P, DepsT, FormattableT]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ContextResponse[DepsT, FormattableT]: ...

    def __call__(
        self,
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Generates a response using the provided model."""
        return self.call(model, ctx, *args, **kwargs)

    @overload
    def call(
        self: "ContextPrompt[P, DepsT, None]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ContextResponse[DepsT, None]: ...

    @overload
    def call(
        self: "ContextPrompt[P, DepsT, FormattableT]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ContextResponse[DepsT, FormattableT]: ...

    def call(
        self,
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Generates a response using the provided model."""
        if isinstance(model, str):
            model = Model(model)
        messages = self.messages(ctx, *args, **kwargs)
        return model.context_call(
            messages, ctx=ctx, tools=self.toolkit, format=self.format
        )

    @overload
    def stream(
        self: "ContextPrompt[P, DepsT, None]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ContextStreamResponse[DepsT, None]: ...

    @overload
    def stream(
        self: "ContextPrompt[P, DepsT, FormattableT]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ContextStreamResponse[DepsT, FormattableT]: ...

    def stream(
        self,
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Generates a streaming response using the provided model."""
        if isinstance(model, str):
            model = Model(model)
        messages = self.messages(ctx, *args, **kwargs)
        return model.context_stream(
            messages, ctx=ctx, tools=self.toolkit, format=self.format
        )


class AsyncContextPrompt(
    BasePrompt[AsyncContextMessageTemplate[P, DepsT]], Generic[P, DepsT, FormattableT]
):
    """An async context-aware prompt that can be called with a model to generate a response.

    Created by decorating an async `ContextMessageTemplate` with `llm.prompt`. The decorated
    async function (with first parameter `'ctx'` of type `Context[DepsT]`) becomes callable
    with a `Model` to generate LLM responses asynchronously with context dependencies.

    An `AsyncContextPrompt` is essentially: async `ContextMessageTemplate` + tools + format.
    It can be invoked with a model: `await prompt(model, ctx, *args, **kwargs)`.
    """

    toolkit: AsyncContextToolkit[DepsT]
    """The toolkit containing this prompt's async context-aware tools."""

    format: FormatSpec[FormattableT] | None
    """The response format for the generated response."""

    def __init__(
        self,
        *,
        fn: AsyncContextMessageTemplate[P, DepsT],
        toolkit: AsyncContextToolkit[DepsT],
        format: FormatSpec[FormattableT] | None,
    ) -> None:
        self.toolkit = toolkit
        self.format = format
        super().__init__(fn=fn)

    async def messages(
        self, ctx: Context[DepsT], *args: P.args, **kwargs: P.kwargs
    ) -> Sequence[Message]:
        """Return the `Messages` from invoking this prompt."""
        return promote_to_messages(await self.fn(ctx, *args, **kwargs))

    @overload
    async def __call__(
        self: "AsyncContextPrompt[P, DepsT, None]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncContextResponse[DepsT, None]: ...

    @overload
    async def __call__(
        self: "AsyncContextPrompt[P, DepsT, FormattableT]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncContextResponse[DepsT, FormattableT]: ...

    async def __call__(
        self,
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Generates a response using the provided model asynchronously."""
        return await self.call(model, ctx, *args, **kwargs)

    @overload
    async def call(
        self: "AsyncContextPrompt[P, DepsT, None]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncContextResponse[DepsT, None]: ...

    @overload
    async def call(
        self: "AsyncContextPrompt[P, DepsT, FormattableT]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncContextResponse[DepsT, FormattableT]: ...

    async def call(
        self,
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Generates a response using the provided model asynchronously."""
        if isinstance(model, str):
            model = Model(model)
        messages = await self.messages(ctx, *args, **kwargs)
        return await model.context_call_async(
            messages, ctx=ctx, tools=self.toolkit, format=self.format
        )

    @overload
    async def stream(
        self: "AsyncContextPrompt[P, DepsT, None]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncContextStreamResponse[DepsT, None]: ...

    @overload
    async def stream(
        self: "AsyncContextPrompt[P, DepsT, FormattableT]",
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncContextStreamResponse[DepsT, FormattableT]: ...

    async def stream(
        self,
        model: Model | ModelId,
        ctx: Context[DepsT],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Generates a streaming response using the provided model asynchronously."""
        if isinstance(model, str):
            model = Model(model)
        messages = await self.messages(ctx, *args, **kwargs)
        return await model.context_stream_async(
            messages, ctx=ctx, tools=self.toolkit, format=self.format
        )



"""Ollama provider implementation."""

import os
from typing import ClassVar

from openai import AsyncOpenAI, OpenAI

from ..openai.completions.base_provider import BaseOpenAICompletionsProvider


class OllamaProvider(BaseOpenAICompletionsProvider):
    """Provider for Ollama's OpenAI-compatible API.

    Inherits from BaseOpenAICompletionsProvider with Ollama-specific configuration:
    - Uses Ollama's local API endpoint (default: http://localhost:11434/v1/)
    - API key is not required (Ollama ignores API keys)
    - Supports OLLAMA_BASE_URL environment variable

    Usage:
        Register the provider with model ID prefixes you want to use:

        ```python
        import llm

        # Register for ollama models
        llm.register_provider("ollama", "ollama/")

        # Now you can use ollama models directly
        @llm.call("ollama/llama2")
        def my_prompt():
            return [llm.messages.user("Hello!")]
        ```
    """

    id: ClassVar[str] = "ollama"
    default_scope: ClassVar[str | list[str]] = "ollama/"
    default_base_url: ClassVar[str | None] = "http://localhost:11434/v1/"
    api_key_env_var: ClassVar[str] = "OLLAMA_API_KEY"
    api_key_required: ClassVar[bool] = False
    provider_name: ClassVar[str | None] = "Ollama"

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | None = None,
    ) -> None:
        """Initialize the Ollama provider.

        Args:
            api_key: API key (optional). Defaults to OLLAMA_API_KEY env var or 'ollama'.
            base_url: Custom base URL. Defaults to OLLAMA_BASE_URL env var
                or http://localhost:11434/v1/.
        """
        resolved_api_key = api_key or os.environ.get(self.api_key_env_var) or "ollama"
        resolved_base_url = (
            base_url or os.environ.get("OLLAMA_BASE_URL") or self.default_base_url
        )

        self.client = OpenAI(
            api_key=resolved_api_key,
            base_url=resolved_base_url,
        )
        self.async_client = AsyncOpenAI(
            api_key=resolved_api_key,
            base_url=resolved_base_url,
        )

    def _model_name(self, model_id: str) -> str:
        """Strip 'ollama/' prefix from model ID for Ollama API."""
        return model_id.removeprefix("ollama/")



"""Configuration for retry behavior."""

from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypedDict
from typing_extensions import Unpack

from ..exceptions import (
    ConnectionError,
    RateLimitError,
    ServerError,
    TimeoutError,
)

if TYPE_CHECKING:
    from ..models import Model
    from ..providers import ModelId

DEFAULT_RETRYABLE_ERRORS: tuple[type[Exception], ...] = (
    ConnectionError,
    RateLimitError,
    ServerError,
    TimeoutError,
)
"""Default exceptions that trigger a retry.

These are transient errors that are likely to succeed on retry:
- ConnectionError: Network issues, DNS failures
- RateLimitError: Rate limits exceeded (429)
- ServerError: Provider-side errors (500+)
- TimeoutError: Request timeouts
"""

DEFAULT_MAX_RETRIES: int = 3
"""Default maximum number of retries after the initial attempt fails."""

DEFAULT_INITIAL_DELAY: float = 0.5
"""Default initial delay in seconds before the first retry."""

DEFAULT_MAX_DELAY: float = 60.0
"""Default maximum delay in seconds between retries."""

DEFAULT_BACKOFF_MULTIPLIER: float = 2.0
"""Default multiplier for exponential backoff (delay *= multiplier after each retry)."""

DEFAULT_JITTER: float = 0.0
"""Default jitter factor (0.0 to 1.0) to add randomness to delays.

A jitter of 0.1 means +/- 10% random variation on the calculated delay.
"""


class RetryArgs(TypedDict, total=False):
    """Arguments for configuring retry behavior.

    This TypedDict is used for the user-facing API where all fields are optional.
    Use RetryConfig for the internal representation with defaults applied.

    Attributes:
        max_retries: Maximum number of retries after the initial attempt fails.
            Defaults to 3.
        retry_on: Tuple of exception types that should trigger a retry.
            Defaults to DEFAULT_RETRYABLE_ERRORS (ConnectionError, RateLimitError,
            ServerError, TimeoutError).
        initial_delay: Initial delay in seconds before the first retry. Defaults to 0.5.
        max_delay: Maximum delay in seconds between retries. Defaults to 60.0.
        backoff_multiplier: Multiplier for exponential backoff. Defaults to 2.0.
        jitter: Jitter factor (0.0 to 1.0) to add randomness to delays. Defaults to 0.0.
    """

    max_retries: int
    """Maximum number of retries after the initial attempt fails. Defaults to 3."""

    retry_on: tuple[type[Exception], ...]
    """Tuple of exception types that should trigger a retry.

    Defaults to DEFAULT_RETRYABLE_ERRORS (ConnectionError, RateLimitError,
    ServerError, TimeoutError).
    """

    initial_delay: float
    """Initial delay in seconds before the first retry. Defaults to 1.0."""

    max_delay: float
    """Maximum delay in seconds between retries. Defaults to 60.0."""

    backoff_multiplier: float
    """Multiplier for exponential backoff (delay *= multiplier after each retry). Defaults to 2.0."""

    jitter: float
    """Jitter factor (0.0 to 1.0) to add randomness to delays. Defaults to 0.0."""

    fallback_models: "Sequence[Model | ModelId]"
    """Sequence of fallback models to try if the primary model fails.

    Each model gets its own full retry budget (max_retries applies per model).
    ModelId strings inherit params from the primary model; Model instances
    use their own params. Defaults to empty tuple (no fallbacks).
    """


@dataclass(frozen=True)
class RetryConfig:
    """Configuration for retry behavior with defaults applied.

    Attributes:
        max_retries: Maximum number of retries after the initial attempt fails.
        retry_on: Tuple of exception types that should trigger a retry.
        initial_delay: Initial delay in seconds before the first retry.
        max_delay: Maximum delay in seconds between retries.
        backoff_multiplier: Multiplier for exponential backoff.
        jitter: Jitter factor (0.0 to 1.0) to add randomness to delays.
    """

    max_retries: int = DEFAULT_MAX_RETRIES
    """Maximum number of retries after the initial attempt fails."""

    retry_on: tuple[type[Exception], ...] = DEFAULT_RETRYABLE_ERRORS
    """Tuple of exception types that should trigger a retry."""

    fallback_models: "tuple[Model | ModelId, ...]" = ()
    """Sequence of fallback models to try if the primary model fails.

    Each model gets its own full retry budget (max_retries applies per model).
    ModelId strings inherit params from the primary model; Model instances
    use their own params. Defaults to empty tuple (no fallbacks).
    """

    initial_delay: float = DEFAULT_INITIAL_DELAY
    """Initial delay in seconds before the first retry."""

    max_delay: float = DEFAULT_MAX_DELAY
    """Maximum delay in seconds between retries."""

    backoff_multiplier: float = DEFAULT_BACKOFF_MULTIPLIER
    """Multiplier for exponential backoff (delay *= multiplier after each retry)."""

    jitter: float = DEFAULT_JITTER
    """Jitter factor (0.0 to 1.0) to add randomness to delays."""

    def __post_init__(self) -> None:
        """Validate configuration values."""
        if self.max_retries < 0:
            raise ValueError("max_retries must be non-negative")
        if self.initial_delay < 0:
            raise ValueError("initial_delay must be non-negative")
        if self.max_delay < 0:
            raise ValueError("max_delay must be non-negative")
        if self.backoff_multiplier < 1:
            raise ValueError("backoff_multiplier must be >= 1")
        if not 0 <= self.jitter <= 1:
            raise ValueError("jitter must be between 0.0 and 1.0")

    @classmethod
    def from_args(cls, **args: Unpack[RetryArgs]) -> "RetryConfig":
        """Create a RetryConfig from RetryArgs kwargs."""
        fallback_models = args.get("fallback_models", ())
        return cls(
            max_retries=args.get("max_retries", DEFAULT_MAX_RETRIES),
            retry_on=args.get("retry_on", DEFAULT_RETRYABLE_ERRORS),
            fallback_models=tuple(fallback_models),
            initial_delay=args.get("initial_delay", DEFAULT_INITIAL_DELAY),
            max_delay=args.get("max_delay", DEFAULT_MAX_DELAY),
            backoff_multiplier=args.get(
                "backoff_multiplier", DEFAULT_BACKOFF_MULTIPLIER
            ),
            jitter=args.get("jitter", DEFAULT_JITTER),
        )



from mirascope import llm

# Audio from local file
audio = llm.Audio.from_file("recording.mp3")

# Audio from raw bytes
audio_bytes = llm.Audio.from_bytes(b"...")

# Include audio in a message
message = llm.messages.user(["Please transcribe this audio:", audio])



from typing import TypeAlias

from .anthropic import (
    AnthropicModelId,
)
from .google import (
    GoogleModelId,
)
from .mlx import (
    MLXModelId,
)
from .openai import (
    OpenAIModelId,
)

ModelId: TypeAlias = AnthropicModelId | GoogleModelId | OpenAIModelId | MLXModelId | str



from mirascope import llm


@llm.retry(
    fallback_models=[
        "anthropic/claude-3-5-haiku-latest",
        "google/gemini-2.0-flash",
    ]
)
@llm.prompt
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


response = recommend_book("openai/gpt-4o-mini", "fantasy")
print(response.text())
# > The Name of the Wind by Patrick Rothfuss



"""Exporter implementation for OpenTelemetry exporters.

This module provides the export layer for sending OpenTelemetry span
events to the Mirascope ingestion endpoint. It wraps the Fern-generated
Mirascope client to provide the interface needed by OpenTelemetry exporters.
"""

from __future__ import annotations

import logging
import time
from collections.abc import Sequence

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult
from opentelemetry.util.types import AttributeValue

from ....api._generated.traces.types import (
    TracesCreateRequestResourceSpansItem,
    TracesCreateRequestResourceSpansItemResource,
    TracesCreateRequestResourceSpansItemResourceAttributesItem,
    TracesCreateRequestResourceSpansItemResourceAttributesItemValue,
    TracesCreateRequestResourceSpansItemResourceAttributesItemValueArrayValue,
    TracesCreateRequestResourceSpansItemScopeSpansItem,
    TracesCreateRequestResourceSpansItemScopeSpansItemScope,
    TracesCreateRequestResourceSpansItemScopeSpansItemSpansItem,
    TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItem,
    TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValue,
    TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValueArrayValue,
    TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemStatus,
)
from ....api.client import Mirascope
from .utils import to_otlp_any_value

logger = logging.getLogger(__name__)


class MirascopeOTLPExporter(SpanExporter):
    """OTLP/HTTP exporter for completed spans.

    This exporter implements the OpenTelemetry SpanExporter interface
    for exporting completed spans in OTLP format over HTTP. It's
    designed to work with BatchSpanProcessor for efficient batching.

    This uses the Fern auto-generated client for sending converted spans.

    Attributes:
        exporter: Export client for sending events.
        timeout: Request timeout in seconds.
    """

    def __init__(
        self,
        client: Mirascope,
        timeout: float = 30.0,
        max_retry_attempts: int = 3,
    ) -> None:
        """Initialize the telemetry exporter.

        Args:
            client: The Fern-generated Mirascope client instance.
                In the future, this will accept the enhanced client from
                mirascope.api.client that provides error handling and caching
                capabilities.
            timeout: Request timeout in seconds for telemetry operations.
            max_retry_attempts: Maximum number of retry attempts for failed exports.
        """
        self.client = client
        self.timeout = timeout
        self.max_retry_attempts = max_retry_attempts
        self._shutdown = False

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        """Export a batch of spans to the telemetry endpoint.

        This is the standard OpenTelemetry export interface.

        Args:
            spans: Sequence of ReadableSpan objects to export.

        Returns:
            SpanExportResult indicating success or failure.
        """
        if self._shutdown:
            return SpanExportResult.FAILURE

        if not spans:
            return SpanExportResult.SUCCESS

        exceptions: list[Exception] = []
        delay = 0.1

        for i in range(self.max_retry_attempts):
            if i > 0:
                time.sleep(delay)
                delay = min(delay * 2, 5.0)

            try:
                otlp_data = self._convert_spans_to_otlp(spans)
                response = self.client.traces.create(resource_spans=otlp_data)

                if (
                    response
                    and hasattr(response, "partial_success")
                    and response.partial_success
                ):
                    partial_success = response.partial_success
                    if hasattr(partial_success, "rejected_spans"):
                        rejected = partial_success.rejected_spans
                        if rejected is not None and rejected > 0:
                            return SpanExportResult.FAILURE

                return SpanExportResult.SUCCESS

            except Exception as e:
                exceptions.append(e)
                logger.warning(
                    f"Export attempt {i + 1} failed, retrying in {delay}s: {e}"
                )

        logger.error(
            f"Failed to export spans after {self.max_retry_attempts} attempts: {exceptions}"
        )

        return SpanExportResult.FAILURE

    def _convert_spans_to_otlp(
        self, spans: Sequence[ReadableSpan]
    ) -> list[TracesCreateRequestResourceSpansItem]:
        """Convert OpenTelemetry spans to OTLP format.

        Args:
            spans: Sequence of ReadableSpan objects.

        Returns:
            List of ResourceSpans in OTLP format.
        """
        resource_spans_map = {}

        for span in spans:
            try:
                otlp_span = self._convert_span(span)
            except ValueError as e:
                logger.warning(f"Skipping span due to error: {e}")
                continue

            resource_key = id(span.resource) if span.resource else "default"

            if resource_key not in resource_spans_map:
                resource = None
                if span.resource:
                    resource_attrs = []
                    for key, value in span.resource.attributes.items():
                        attr_value = self._convert_resource_attribute_value(value)
                        resource_attrs.append(
                            TracesCreateRequestResourceSpansItemResourceAttributesItem(
                                key=key,
                                value=attr_value,
                            )
                        )
                    resource = TracesCreateRequestResourceSpansItemResource(
                        attributes=resource_attrs
                    )

                resource_spans_map[resource_key] = {
                    "resource": resource,
                    "scope_spans": {},
                }

            scope_key = (
                span.instrumentation_scope.name
                if span.instrumentation_scope
                else "unknown"
            )

            if scope_key not in resource_spans_map[resource_key]["scope_spans"]:
                scope = None
                if span.instrumentation_scope:
                    scope = TracesCreateRequestResourceSpansItemScopeSpansItemScope(
                        name=span.instrumentation_scope.name,
                        version=span.instrumentation_scope.version,
                    )

                resource_spans_map[resource_key]["scope_spans"][scope_key] = {
                    "scope": scope,
                    "spans": [],
                }

            resource_spans_map[resource_key]["scope_spans"][scope_key]["spans"].append(
                otlp_span
            )

        result = []
        for resource_data in resource_spans_map.values():
            scope_spans = []
            for scope_data in resource_data["scope_spans"].values():
                scope_spans.append(
                    TracesCreateRequestResourceSpansItemScopeSpansItem(
                        scope=scope_data["scope"],
                        spans=scope_data["spans"],
                    )
                )

            result.append(
                TracesCreateRequestResourceSpansItem(
                    resource=resource_data["resource"],
                    scope_spans=scope_spans,
                )
            )

        return result

    def _convert_span(
        self, span: ReadableSpan
    ) -> TracesCreateRequestResourceSpansItemScopeSpansItemSpansItem:
        """Convert a single ReadableSpan to OTLP format."""
        context = span.get_span_context()
        if not context or not context.is_valid:
            raise ValueError(f"Cannot export span without valid context: {span.name}")

        attributes = []
        if span.attributes:
            for key, value in span.attributes.items():
                attr_value = self._convert_attribute_value(value)
                attributes.append(
                    TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItem(
                        key=key,
                        value=attr_value,
                    )
                )

        status = None
        if span.status:
            status = TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemStatus(
                code=span.status.status_code.value,
                message=span.status.description or "",
            )

        # Convert events
        events = None
        if span.events:
            events = []
            for event in span.events:
                event_attrs: list[dict[str, object]] = []
                if event.attributes:
                    for key, value in event.attributes.items():
                        # Convert to OTLP attribute format with typed values
                        attr_value = self._convert_event_attribute_value(value)
                        event_attrs.append({"key": key, "value": attr_value})
                events.append(
                    {
                        "name": event.name,
                        "timeUnixNano": str(event.timestamp)
                        if event.timestamp
                        else None,
                        "attributes": event_attrs if event_attrs else None,
                    }
                )

        trace_id = format(context.trace_id, "032x")
        span_id = format(context.span_id, "016x")

        # Build kwargs, only including events if present (to avoid sending null)
        kwargs: dict[str, object] = {
            "trace_id": trace_id,
            "span_id": span_id,
            "parent_span_id": (
                format(span.parent.span_id, "016x")
                if span.parent and span.parent.span_id
                else None
            ),
            "name": span.name,
            "kind": span.kind.value if span.kind else 0,
            "start_time_unix_nano": str(span.start_time) if span.start_time else "0",
            "end_time_unix_nano": str(span.end_time) if span.end_time else "0",
            "attributes": attributes or None,
            "status": status,
        }
        # Only include events if present (omit entirely to avoid sending null)
        if events:
            kwargs["events"] = events

        return TracesCreateRequestResourceSpansItemScopeSpansItemSpansItem(**kwargs)  # type: ignore[arg-type]

    def _convert_attribute_value(
        self, value: AttributeValue
    ) -> TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValue:
        """Convert OpenTelemetry AttributeValue to Mirascope API's KeyValueValue."""
        match value:
            case str():
                return TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValue(
                    string_value=value
                )
            case bool():
                return TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValue(
                    bool_value=value
                )
            case int():
                return TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValue(
                    int_value=str(value)
                )
            case float():
                return TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValue(
                    double_value=value
                )
            case _:
                return TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValue(
                    array_value=TracesCreateRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValueArrayValue(
                        values=[to_otlp_any_value(v) for v in value]
                    )
                )

    def _convert_event_attribute_value(
        self, value: AttributeValue
    ) -> dict[str, object]:
        """Convert OpenTelemetry AttributeValue to OTLP event attribute value format."""
        return to_otlp_any_value(value)

    def _convert_resource_attribute_value(
        self, value: AttributeValue
    ) -> TracesCreateRequestResourceSpansItemResourceAttributesItemValue:
        """Convert OpenTelemetry AttributeValue to Mirascope API's resource KeyValueValue."""
        match value:
            case str():
                return TracesCreateRequestResourceSpansItemResourceAttributesItemValue(
                    string_value=value
                )
            case bool():
                return TracesCreateRequestResourceSpansItemResourceAttributesItemValue(
                    bool_value=value
                )
            case int():
                return TracesCreateRequestResourceSpansItemResourceAttributesItemValue(
                    int_value=str(value)
                )
            case float():
                return TracesCreateRequestResourceSpansItemResourceAttributesItemValue(
                    double_value=value
                )
            case _:
                return TracesCreateRequestResourceSpansItemResourceAttributesItemValue(
                    array_value=TracesCreateRequestResourceSpansItemResourceAttributesItemValueArrayValue(
                        values=[to_otlp_any_value(v) for v in value]
                    )
                )

    def shutdown(self) -> None:
        """Shutdown the exporter. Subsequent exports will return FAILURE."""
        self._shutdown = True

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        """Force flush any pending data.

        No-op since this exporter does not buffer data internally.

        Args:
            timeout_millis: Maximum time to wait in milliseconds (unused).

        Returns:
            Always True since there is no internal buffer to flush.
        """
        return True



from mirascope import llm


@llm.prompt
def recommend_book(genre: str):
    return f"Please recommend a book in {genre}."


response = recommend_book("anthropic/claude-haiku-4-5", "fantasy")
print(response.text())



"""Retry response classes that extend base responses with retry metadata."""

from typing import TYPE_CHECKING, Generic, overload

from ..context import Context, DepsT
from ..formatting import FormattableT
from ..messages import UserContent
from ..responses import (
    AsyncContextResponse,
    AsyncResponse,
    ContextResponse,
    Response,
)
from .utils import RetryFailure, get_retry_model_from_context

if TYPE_CHECKING:
    from .retry_config import RetryConfig
    from .retry_models import RetryModel


class RetryResponse(Response[FormattableT]):
    """Response that includes retry metadata.

    Extends `Response` directly, copying all attributes from a wrapped response
    and adding retry configuration. The `model` property returns a `RetryModel`
    with the successful model as the active model.
    """

    _retry_model: "RetryModel"
    """The RetryModel with the successful model as active."""

    retry_failures: list[RetryFailure]
    """Failed attempts before the successful one (empty if first attempt succeeded)."""

    def __init__(
        self,
        response: Response[FormattableT],
        retry_model: "RetryModel",
        retry_failures: list[RetryFailure],
    ) -> None:
        """Initialize a RetryResponse.

        Args:
            response: The successful response from the LLM.
            retry_model: RetryModel with the successful model as active.
            retry_failures: List of failed attempts before success.
        """
        # Copy all attributes from the wrapped response
        for key, value in response.__dict__.items():
            object.__setattr__(self, key, value)
        self._retry_model = retry_model
        self.retry_failures = retry_failures

    @property
    def model(self) -> "RetryModel":
        """A RetryModel with the successful model as active.

        If a model is set in context (via `llm.model()` or `llm.retry_model()`),
        that model is used instead, wrapped with this response's retry config.

        Otherwise, this RetryModel has the same pool of available models, but with
        the model that succeeded set as the active model. This means:
        - `response.model.model_id` equals `response.model_id`
        - `response.resume()` will try the successful model first
        """
        return get_retry_model_from_context(self._retry_model)

    @property
    def retry_config(self) -> "RetryConfig":
        """The retry configuration for this response."""
        return self.model.retry_config

    @overload
    def resume(
        self: "RetryResponse[None]", content: UserContent
    ) -> "RetryResponse[None]": ...

    @overload
    def resume(
        self: "RetryResponse[FormattableT]", content: UserContent
    ) -> "RetryResponse[FormattableT]": ...

    def resume(
        self, content: UserContent
    ) -> "RetryResponse[None] | RetryResponse[FormattableT]":
        """Generate a new RetryResponse using this response's messages with additional user content.

        Args:
            content: The new user message content to append to the message history.

        Returns:
            A new RetryResponse instance generated from the extended message history.
        """
        return self.model.resume(response=self, content=content)


class AsyncRetryResponse(AsyncResponse[FormattableT]):
    """Async response that includes retry metadata.

    Extends `AsyncResponse` directly, copying all attributes from a wrapped response
    and adding retry configuration. The `model` property returns a `RetryModel`
    with the successful model as the active model.
    """

    _retry_model: "RetryModel"
    """The RetryModel with the successful model as active."""

    retry_failures: list[RetryFailure]
    """Failed attempts before the successful one (empty if first attempt succeeded)."""

    def __init__(
        self,
        response: AsyncResponse[FormattableT],
        retry_model: "RetryModel",
        retry_failures: list[RetryFailure],
    ) -> None:
        """Initialize an AsyncRetryResponse.

        Args:
            response: The successful async response from the LLM.
            retry_model: RetryModel with the successful model as active.
            retry_failures: List of failed attempts before success.
        """
        # Copy all attributes from the wrapped response
        for key, value in response.__dict__.items():
            object.__setattr__(self, key, value)
        self._retry_model = retry_model
        self.retry_failures = retry_failures

    @property
    def model(self) -> "RetryModel":
        """A RetryModel with the successful model as active.

        If a model is set in context (via `llm.model()` or `llm.retry_model()`),
        that model is used instead, wrapped with this response's retry config.

        Otherwise, this RetryModel has the same pool of available models, but with
        the model that succeeded set as the active model. This means:
        - `response.model.model_id` equals `response.model_id`
        - `response.resume()` will try the successful model first
        """
        return get_retry_model_from_context(self._retry_model)

    @property
    def retry_config(self) -> "RetryConfig":
        """The retry configuration for this response."""
        return self.model.retry_config

    @overload
    async def resume(
        self: "AsyncRetryResponse[None]", content: UserContent
    ) -> "AsyncRetryResponse[None]": ...

    @overload
    async def resume(
        self: "AsyncRetryResponse[FormattableT]", content: UserContent
    ) -> "AsyncRetryResponse[FormattableT]": ...

    async def resume(
        self, content: UserContent
    ) -> "AsyncRetryResponse[None] | AsyncRetryResponse[FormattableT]":
        """Generate a new AsyncRetryResponse using this response's messages with additional user content.

        Args:
            content: The new user message content to append to the message history.

        Returns:
            A new AsyncRetryResponse instance generated from the extended message history.
        """
        return await self.model.resume_async(response=self, content=content)


class ContextRetryResponse(
    ContextResponse[DepsT, FormattableT], Generic[DepsT, FormattableT]
):
    """Context response that includes retry metadata.

    Extends `ContextResponse` directly, copying all attributes from a wrapped response
    and adding retry configuration. The `model` property returns a `RetryModel`
    with the successful model as the active model.
    """

    _retry_model: "RetryModel"
    """The RetryModel with the successful model as active."""

    retry_failures: list[RetryFailure]
    """Failed attempts before the successful one (empty if first attempt succeeded)."""

    def __init__(
        self,
        response: ContextResponse[DepsT, FormattableT],
        retry_model: "RetryModel",
        retry_failures: list[RetryFailure],
    ) -> None:
        """Initialize a ContextRetryResponse.

        Args:
            response: The successful context response from the LLM.
            retry_model: RetryModel with the successful model as active.
            retry_failures: List of failed attempts before success.
        """
        # Copy all attributes from the wrapped response
        for key, value in response.__dict__.items():
            object.__setattr__(self, key, value)
        self._retry_model = retry_model
        self.retry_failures = retry_failures

    @property
    def model(self) -> "RetryModel":
        """A RetryModel with the successful model as active.

        If a model is set in context (via `llm.model()` or `llm.retry_model()`),
        that model is used instead, wrapped with this response's retry config.

        Otherwise, this RetryModel has the same pool of available models, but with
        the model that succeeded set as the active model. This means:
        - `response.model.model_id` equals `response.model_id`
        - `response.resume()` will try the successful model first
        """
        return get_retry_model_from_context(self._retry_model)

    @property
    def retry_config(self) -> "RetryConfig":
        """The retry configuration for this response."""
        return self.model.retry_config

    @overload
    def resume(
        self: "ContextRetryResponse[DepsT, None]",
        ctx: Context[DepsT],
        content: UserContent,
    ) -> "ContextRetryResponse[DepsT, None]": ...

    @overload
    def resume(
        self: "ContextRetryResponse[DepsT, FormattableT]",
        ctx: Context[DepsT],
        content: UserContent,
    ) -> "ContextRetryResponse[DepsT, FormattableT]": ...

    def resume(
        self, ctx: Context[DepsT], content: UserContent
    ) -> (
        "ContextRetryResponse[DepsT, None] | ContextRetryResponse[DepsT, FormattableT]"
    ):
        """Generate a new ContextRetryResponse using this response's messages with additional user content.

        Args:
            ctx: A `Context` with the required deps type.
            content: The new user message content to append to the message history.

        Returns:
            A new ContextRetryResponse instance generated from the extended message history.
        """
        return self.model.context_resume(ctx=ctx, response=self, content=content)


class AsyncContextRetryResponse(
    AsyncContextResponse[DepsT, FormattableT], Generic[DepsT, FormattableT]
):
    """Async context response that includes retry metadata.

    Extends `AsyncContextResponse` directly, copying all attributes from a wrapped response
    and adding retry configuration. The `model` property returns a `RetryModel`
    with the successful model as the active model.
    """

    _retry_model: "RetryModel"
    """The RetryModel with the successful model as active."""

    retry_failures: list[RetryFailure]
    """Failed attempts before the successful one (empty if first attempt succeeded)."""

    def __init__(
        self,
        response: AsyncContextResponse[DepsT, FormattableT],
        retry_model: "RetryModel",
        retry_failures: list[RetryFailure],
    ) -> None:
        """Initialize an AsyncContextRetryResponse.

        Args:
            response: The successful async context response from the LLM.
            retry_model: RetryModel with the successful model as active.
            retry_failures: List of failed attempts before success.
        """
        # Copy all attributes from the wrapped response
        for key, value in response.__dict__.items():
            object.__setattr__(self, key, value)
        self._retry_model = retry_model
        self.retry_failures = retry_failures

    @property
    def model(self) -> "RetryModel":
        """A RetryModel with the successful model as active.

        If a model is set in context (via `llm.model()` or `llm.retry_model()`),
        that model is used instead, wrapped with this response's retry config.

        Otherwise, this RetryModel has the same pool of available models, but with
        the model that succeeded set as the active model. This means:
        - `response.model.model_id` equals `response.model_id`
        - `response.resume()` will try the successful model first
        """
        return get_retry_model_from_context(self._retry_model)

    @property
    def retry_config(self) -> "RetryConfig":
        """The retry configuration for this response."""
        return self.model.retry_config

    @overload
    async def resume(
        self: "AsyncContextRetryResponse[DepsT, None]",
        ctx: Context[DepsT],
        content: UserContent,
    ) -> "AsyncContextRetryResponse[DepsT, None]": ...

    @overload
    async def resume(
        self: "AsyncContextRetryResponse[DepsT, FormattableT]",
        ctx: Context[DepsT],
        content: UserContent,
    ) -> "AsyncContextRetryResponse[DepsT, FormattableT]": ...

    async def resume(
        self, ctx: Context[DepsT], content: UserContent
    ) -> "AsyncContextRetryResponse[DepsT, None] | AsyncContextRetryResponse[DepsT, FormattableT]":
        """Generate a new AsyncContextRetryResponse using this response's messages with additional user content.

        Args:
            ctx: A `Context` with the required deps type.
            content: The new user message content to append to the message history.

        Returns:
            A new AsyncContextRetryResponse instance generated from the extended message history.
        """
        return await self.model.context_resume_async(
            ctx=ctx, response=self, content=content
        )



"""End-to-end tests for Mirascope Router provider."""

import os

import pytest

from mirascope import llm
from tests.utils import (
    Snapshot,
    snapshot_test,
)

MIRASCOPE_MODEL_IDS = [
    "openai/gpt-5-mini",
    "openai/gpt-5-mini:completions",
    "openai/gpt-5-mini:responses",
    "anthropic/claude-haiku-4-5",
    "google/gemini-2.5-flash",
]


@pytest.mark.parametrize("model_id", MIRASCOPE_MODEL_IDS)
@pytest.mark.vcr
def test_mirascope_provider(model_id: llm.ModelId, snapshot: Snapshot) -> None:
    """Test that Mirascope Router provider works correctly."""

    @llm.call(model_id)
    def add_numbers(a: int, b: int) -> str:
        return f"What is {a} + {b}?"

    llm.register_provider(
        "mirascope",
        base_url="http://localhost:3000/router/v2",
    )

    model_scope, model_name = model_id.split("/", 1)

    with snapshot_test(snapshot) as snap:
        response = add_numbers(4200, 42)
        assert response.provider_id == model_scope
        assert model_name in response.provider_model_name
        snap.set_response(response)
        assert "4242" in response.pretty(), (
            f"Expected '4242' in response: {response.pretty()}"
        )


@pytest.mark.parametrize("model_id", MIRASCOPE_MODEL_IDS)
@pytest.mark.vcr
def test_mirascope_provider_streaming(
    model_id: llm.ModelId, snapshot: Snapshot
) -> None:
    """Test that Mirascope Router provider works correctly with streaming."""

    @llm.call(model_id)
    def add_numbers(a: int, b: int) -> str:
        return f"What is {a} + {b}?"

    llm.register_provider(
        "mirascope",
        base_url="http://localhost:3000/router/v2",
    )

    model_scope, model_name = model_id.split("/", 1)

    with snapshot_test(snapshot) as snap:
        stream = add_numbers.stream(4200, 42)
        assert stream.provider_id == model_scope
        assert model_name in stream.provider_model_name

        stream.finish()
        content = stream.pretty()

        snap.set_response(stream)
        assert "4242" in content, f"Expected '4242' in streamed content: {content}"


def test_mirascope_provider_missing_api_key() -> None:
    """Test that Mirascope provider raises clear error when API key is missing."""
    original_key = os.environ.pop("MIRASCOPE_API_KEY", None)
    try:
        with pytest.raises(ValueError) as exc_info:
            llm.providers.MirascopeProvider()
        assert "Mirascope API key not found" in str(exc_info.value)
        assert "MIRASCOPE_API_KEY" in str(exc_info.value)
    finally:
        if original_key is not None:
            os.environ["MIRASCOPE_API_KEY"] = original_key



from dataclasses import dataclass

from mirascope import llm


@dataclass
class Library:
    books: dict[str, str]  # title -> author


@llm.tool
def get_author(ctx: llm.Context[Library], title: str) -> str:
    """Get the author of a book."""
    return ctx.deps.books.get(title, "Book not found")


library = Library(books={"Dune": "Frank Herbert", "Neuromancer": "William Gibson"})
ctx = llm.Context(deps=library)
model = llm.use_model("openai/gpt-5-mini")
response = model.context_call(
    "Who wrote Dune?",
    ctx=ctx,
    tools=[get_author],
)

while response.tool_calls:
    tool_outputs = response.execute_tools(ctx)
    response = response.resume(ctx, tool_outputs)

print(response.pretty())
# Dune was written by Frank Herbert.



"""Tests for the `format` function."""

from typing import Literal

import pytest
from inline_snapshot import snapshot
from pydantic import BaseModel

from mirascope import llm
from tests import utils


@pytest.mark.parametrize("mode", ["strict", "json", "tool"])
def test_format_all_modes(mode: llm.FormattingMode) -> None:
    """Test `format` function with all supported modes."""

    class ModeTest(BaseModel):
        """Basic test of a BaseModel derived Formattable"""

        content: str

    format = llm.format(ModeTest, mode=mode)
    assert (
        utils.format_snapshot(format)
        == snapshot(
            {
                "strict": {
                    "name": "ModeTest",
                    "description": "Basic test of a BaseModel derived Formattable",
                    "schema": {
                        "description": "Basic test of a BaseModel derived Formattable",
                        "properties": {
                            "content": {"title": "Content", "type": "string"}
                        },
                        "required": ["content"],
                        "title": "ModeTest",
                        "type": "object",
                    },
                    "mode": "strict",
                    "formatting_instructions": None,
                },
                "json": {
                    "name": "ModeTest",
                    "description": "Basic test of a BaseModel derived Formattable",
                    "schema": {
                        "description": "Basic test of a BaseModel derived Formattable",
                        "properties": {
                            "content": {"title": "Content", "type": "string"}
                        },
                        "required": ["content"],
                        "title": "ModeTest",
                        "type": "object",
                    },
                    "mode": "json",
                    "formatting_instructions": """\
Respond only with valid JSON that matches this exact schema:
{
  "description": "Basic test of a BaseModel derived Formattable",
  "properties": {
    "content": {
      "title": "Content",
      "type": "string"
    }
  },
  "required": [
    "content"
  ],
  "title": "ModeTest",
  "type": "object"
}\
""",
                },
                "tool": {
                    "name": "ModeTest",
                    "description": "Basic test of a BaseModel derived Formattable",
                    "schema": {
                        "description": "Basic test of a BaseModel derived Formattable",
                        "properties": {
                            "content": {"title": "Content", "type": "string"}
                        },
                        "required": ["content"],
                        "title": "ModeTest",
                        "type": "object",
                    },
                    "mode": "tool",
                    "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
                },
            }
        )[mode]
    )


def test_format_empty_model() -> None:
    """Test `format` function with all supported modes."""

    class Empty(BaseModel):
        pass

    format = llm.format(Empty, mode="tool")
    assert utils.format_snapshot(format) == snapshot(
        {
            "name": "Empty",
            "description": None,
            "schema": {"properties": {}, "title": "Empty", "type": "object"},
            "mode": "tool",
            "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
        }
    )


def test_format_nested_models() -> None:
    """Test `format` function with nested Pydantic models."""

    class Address(BaseModel):
        street: str
        city: str

    class Person(BaseModel):
        name: str
        address: Address

    format = llm.format(Person, mode="tool")

    assert utils.format_snapshot(format) == snapshot(
        {
            "name": "Person",
            "description": None,
            "schema": {
                "$defs": {
                    "Address": {
                        "properties": {
                            "street": {"title": "Street", "type": "string"},
                            "city": {"title": "City", "type": "string"},
                        },
                        "required": ["street", "city"],
                        "title": "Address",
                        "type": "object",
                    }
                },
                "properties": {
                    "name": {"title": "Name", "type": "string"},
                    "address": {"$ref": "#/$defs/Address"},
                },
                "required": ["name", "address"],
                "title": "Person",
                "type": "object",
            },
            "mode": "tool",
            "formatting_instructions": "Always respond to the user's query using the __mirascope_formatted_output_tool__ tool for structured output.",
        }
    )


def test_format_primitive_str() -> None:
    """Test `format` function with str primitive."""
    format = llm.format(str, mode="json")

    assert format is not None
    assert format.name == "str"
    assert format.description is None  # Primitives don't have docstrings
    assert format.mode == "json"
    assert "output" in format.schema["properties"]  # pyright: ignore[reportOperatorIssue]
    assert format.schema["properties"]["output"]["type"] == "string"  # pyright: ignore[reportIndexIssue]


def test_format_primitive_int() -> None:
    """Test `format` function with int primitive."""
    format = llm.format(int, mode="strict")

    assert format is not None
    assert format.name == "int"
    assert format.mode == "strict"
    assert "output" in format.schema["properties"]  # pyright: ignore[reportOperatorIssue]
    assert format.schema["properties"]["output"]["type"] == "integer"  # pyright: ignore[reportIndexIssue]


def test_format_list_of_models() -> None:
    """Test `format` function with list[Model]."""

    class Book(BaseModel):
        """A book."""

        title: str
        author: str

    format = llm.format(list[Book], mode="tool")

    assert format is not None
    assert format.mode == "tool"
    assert "output" in format.schema["properties"]  # pyright: ignore[reportOperatorIssue]

    # Should have Book definition in $defs
    assert "$defs" in format.schema or "definitions" in format.schema
    schema_defs = format.schema.get("$defs", format.schema.get("definitions", {}))
    assert "Book" in schema_defs  # pyright: ignore[reportOperatorIssue]


def test_format_dict_primitive() -> None:
    """Test `format` function with dict[str, int]."""
    format = llm.format(dict[str, int], mode="json")

    assert format is not None
    assert "output" in format.schema["properties"]  # pyright: ignore[reportOperatorIssue]
    # Dict schema should have additionalProperties with type integer
    output_schema = format.schema["properties"]["output"]  # pyright: ignore[reportIndexIssue]
    assert output_schema["type"] == "object"


def test_format_union_type() -> None:
    """Test `format` function with Union type."""
    format = llm.format(str | int, mode="json")  # pyright: ignore[reportArgumentType]

    assert format is not None
    assert "output" in format.schema["properties"]  # pyright: ignore[reportOperatorIssue]


def test_format_literal_type() -> None:
    """Test `format` function with Literal type."""
    format = llm.format(Literal["a", "b", "c"], mode="json")  # pyright: ignore[reportArgumentType]

    assert format is not None
    assert "output" in format.schema["properties"]  # pyright: ignore[reportOperatorIssue]


def test_format_none_returns_none() -> None:
    """Test that format(None) returns None."""
    assert llm.format(None, mode="json") is None


def test_format_with_output_parser() -> None:
    """Test `format` function with an OutputParser."""

    @llm.output_parser(formatting_instructions="Return XML format")
    def parse_xml(response: llm.AnyResponse) -> str:
        """Parse XML response."""
        return "parsed"

    format = llm.format(parse_xml, mode="parser")

    assert format is not None
    assert format.name == "parse_xml"
    assert format.description == "Parse XML response."
    assert format.schema == {}
    assert format.mode == "parser"
    assert format.formattable is parse_xml
    assert format.formatting_instructions == "Return XML format"


def test_format_output_parser_wrong_mode_raises_error() -> None:
    """Test that format() raises ValueError when mode is not 'parser' for OutputParser."""

    @llm.output_parser(formatting_instructions="Return XML format")
    def parse_xml(response: llm.AnyResponse) -> str:
        """Parse XML response."""
        return "parsed"

    with pytest.raises(
        ValueError,
        match="mode must be 'parser' for OutputParser, got 'json'",
    ):
        llm.format(parse_xml, mode="json")


def test_resolve_format_with_output_parser() -> None:
    """Test `resolve_format` function with an OutputParser."""

    @llm.output_parser(formatting_instructions="Return CSV format")
    def parse_csv(response: llm.AnyResponse) -> list[str]:
        """Parse CSV response."""
        return []

    format = llm.formatting.resolve_format(parse_csv, default_mode="json")

    assert format is not None
    assert format.name == "parse_csv"
    assert format.description == "Parse CSV response."
    assert format.schema == {}
    assert format.mode == "parser"
    assert format.formattable is parse_csv



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ...core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .traces_create_otel_request_resource_spans_item_scope_spans_item_spans_item_attributes_item_value import (
    TracesCreateOtelRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValue,
)


class TracesCreateOtelRequestResourceSpansItemScopeSpansItemSpansItemAttributesItem(
    UniversalBaseModel
):
    key: str
    value: TracesCreateOtelRequestResourceSpansItemScopeSpansItemSpansItemAttributesItemValue

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
            extra="allow", frozen=True
        )  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from mirascope import llm


@llm.retry()
@llm.prompt
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


response = recommend_book("openai/gpt-4o-mini", "fantasy")
print(response.text())
# > The Name of the Wind by Patrick Rothfuss
